--- linux-2.6.27.6/drivers/video/sm501fb.c	2008-11-14 01:56:21.000000000 +0800
+++ linux-2.6.27.6.yaffs/drivers/video/sm501fb.c	2011-05-08 18:00:52.000000000 +0800
@@ -1,16 +1,27 @@
 /* linux/drivers/video/sm501fb.c
  *
+ * Copyright (C) 2010 SiliconMotion Inc.
  * Copyright (c) 2006 Simtec Electronics
  *	Vincent Sanders <vince@simtec.co.uk>
  *	Ben Dooks <ben@simtec.co.uk>
+ *	Boyod.yang,  <boyod.yang@siliconmotion.com.cn>
+ *    Teddy.wang,  <teddy.wang@siliconmotion.com>
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation.
+ * it under the terms of the GNU General Public
+ * License version 2 or later as published by the Free Software Foundation.  
  *
- * Framebuffer driver for the Silicon Motion SM501
+ * Framebuffer driver for the Silicon Motion SM50X/SM107
  */
 
+// #define DEBUG
+
+#ifdef DEBUG
+#define smdbg(format, arg...)	printk(KERN_DEBUG format , ## arg)
+#else
+#define smdbg(format, arg...)
+#endif
+
 #include <linux/module.h>
 #include <linux/kernel.h>
 #include <linux/errno.h>
@@ -38,25 +49,221 @@
 #include <linux/pm.h>
 #endif
 
+#include <linux/screen_info.h>
+
 #include <linux/sm501.h>
-#include <linux/sm501-regs.h>
+#include <linux/sm501_regs.h>
+
+#define SM501_MEMF_CURSOR		(1)
+#define SM501_MEMF_PANEL		(2)
+#define SM501_MEMF_CRT			(4)
+#define SM501_MEMF_ACCEL		(8)
+
+#define DEFAULT_MODE "1368x768-16@60"
+
+
+/*
+ * globals
+ */
+ 
+static char *mode_option __devinitdata = DEFAULT_MODE;
+static int noaccel = 0;
+static int nofr = 0;
+static int noca = 0;
+static int noib = 0;
+static int mirror = 0;
+static int dcolor = 0;
+static int nomtrr = 1;
+static int gen = 0;
+
+typedef enum _spolarity_t
+{
+    POS, /* positive */
+    NEG, /* negative */
+}
+spolarity_t;
+
+typedef struct _mode_parameter_t
+{
+    /* Horizontal timing. */
+    unsigned long horizontal_total;
+    unsigned long horizontal_display_end;
+    unsigned long horizontal_sync_start;
+    unsigned long horizontal_sync_width;
+    spolarity_t horizontal_sync_polarity;
+
+    /* Vertical timing. */
+    unsigned long vertical_total;
+    unsigned long vertical_display_end;
+    unsigned long vertical_sync_start;
+    unsigned long vertical_sync_height;
+    spolarity_t vertical_sync_polarity;
+
+    /* Refresh timing. */
+    unsigned long pixel_clock;
+    unsigned long horizontal_frequency;
+    unsigned long vertical_frequency;
+    
+    /* Clock Phase. This clock phase only applies to Panel. */
+    spolarity_t clock_phase_polarity;
+}mode_parameter_t;
+
+#include "sm2d.c"
+
+/*
+ *  Timing parameter for some popular modes.
+ *  Note that this table is made according to standard VESA parameters for the
+ *  popular modes.
+ */
+static mode_parameter_t gModeParamTable[] =
+{
+/* 320 x 240 */
+ { 352, 320, 335,  8, NEG, 265, 240, 254, 2, NEG,  5600000, 15909, 60, NEG},
+
+/* 400 x 300 */
+ { 528, 400, 420, 64, NEG, 314, 300, 301, 2, NEG,  9960000, 18864, 60, NEG},
+
+/* 640 x 480  [4:3] */
+ { 800, 640, 656, 96, NEG, 525, 480, 490, 2, NEG, 25175000, 31500, 60, NEG},
+ { 840, 640, 680, 64, NEG, 500, 480, 481, 3, NEG, 31500000, 37500, 75, NEG},
+ { 832, 640, 700, 56, NEG, 509, 480, 481, 3, NEG, 36000000, 43269, 85, NEG},
+
+/* 720 x 480  [3:2] */
+ { 889, 720, 738,108, POS, 525, 480, 490, 2, NEG, 28000000, 31496, 60, NEG},
+
+/* 720 x 540  [4:3] */
+ { 886, 720, 740, 96, POS, 576, 540, 545, 2, POS, 30600000, 34537, 60, NEG},
+
+/* 800 x 480  [5:3] */
+ { 973, 800, 822, 56, POS, 524, 480, 490, 2, NEG, 30600000, 31449, 60, NEG},
+
+/* 800 x 600  [4:3] */
+ {1062, 800, 840,128, POS, 628, 600, 601, 4, POS, 40000000, 37665, 60, NEG},
+ {1056, 800, 816, 80, POS, 625, 600, 601, 3, POS, 49500000, 46875, 75, NEG},
+ {1054, 800, 842, 64, POS, 625, 600, 601, 3, POS, 56000000, 53131, 85, NEG},
+
+/* 960 x 720  [4:3] */
+ {1245, 960, 992, 64, POS, 750, 720, 721, 3, POS, 56000000, 44980, 60, NEG},
+      
+/* 1024 x 600  [16:9] 1.7 */
+ {1313,1024,1064,104, POS, 622, 600, 601, 3, POS, 49000000, 37319, 60, NEG},
+     
+/* 1024 x 768  [4:3] */
+ {1340,1024,1060,136, NEG, 809, 768, 772, 6, NEG, 65000000, 48507, 60, NEG},
+ {1337,1024,1072, 96, NEG, 808, 768, 780, 3, NEG, 81000000, 60583, 75, POS},
+ {1376,1024,1070, 96, POS, 808, 768, 769, 3, POS, 94500000, 68677, 85, POS},
+  
+/* 1152 x 864  [4:3] */
+ {1475,1152,1208, 96, NEG, 888, 864, 866, 3, NEG, 78600000, 53288, 60, POS},
+
+/* 1280 x 768  [5:3] */
+ {1678,1280,1350,136, POS, 795, 768, 769, 3, POS, 80000000, 47676, 60, POS},
+
+/* 1280 x 800  [8:5] */
+ {1650,1280,1344,136, NEG, 824, 800, 800, 3, NEG, 81600000, 49455, 60, POS},
+
+/* 1280 x 960  [4:3] */
+ {1618,1280,1330, 96, NEG, 977, 960, 960, 2, NEG, 94500000, 59259, 60, POS},
+    
+/* 1280 x 1024 [5:4] */
+ {1626,1280,1332,112, POS,1046,1024,1024, 2, POS,102000000, 62731, 60, POS},
+
+/* 1366 x 768  [16:9] */
+ {1722,1366,1424,112, NEG, 784, 768, 769, 3, NEG, 81000000, 47038, 60, POS},
+
+/* 1600 x 1200 [4:3] */
+/* Not supported yet.
+ {2160,1600,1724,112, NEG,1220,1200,1200, 2, NEG,158000000, 73148, 60, POS},
+ */
+
+/* End of table. */
+ { 0, 0, 0, 0, NEG, 0, 0, 0, 0, NEG, 0, 0, 0, NEG},
+};
+
+static struct fb_var_screeninfo __devinitdata sm50xfb_default_var = {
+	.xres = 640,
+	.yres = 480,
+	.xres_virtual = 640,
+	.yres_virtual = 480,
+	.bits_per_pixel = 8,
+	.red = {0, 8, 0},
+	.green = {0, 8, 0},
+	.blue = {0, 8, 0},
+	.transp = {0, 0, 0},
+	.activate = FB_ACTIVATE_NOW,
+	.height = -1,
+	.width = -1,
+	.pixclock = 39721,
+	.left_margin = 40,
+	.right_margin = 24,
+	.upper_margin = 32,
+	.lower_margin = 11,
+	.hsync_len = 96,
+	.vsync_len = 2,
+	.vmode = FB_VMODE_NONINTERLACED
+};
+
+
+static const struct fb_videomode SM501modedb[] = {	
+	/* 10 800x600-75 VESA */  /*added by caesar*/
+	{ NULL, 75, 800, 600, 20202, 160, 16, 21, 01, 80, 3,
+	  FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
+	  FB_VMODE_NONINTERLACED, FB_MODE_IS_VESA },
+	/* 15 1024x768-75 VESA */	/*added by caesar*/
+	{ NULL, 75, 1024, 768, 12690, 176, 16, 28, 2, 96, 3,
+	  FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
+	  FB_VMODE_NONINTERLACED, FB_MODE_IS_VESA },
+	/* 16 1024x768-85 VESA */
+	{ NULL, 85, 1024, 768, 10582, 208, 48, 36, 1, 96, 3,
+	  FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
+	  FB_VMODE_NONINTERLACED, FB_MODE_IS_VESA },
+	  /* 20 1280x1024-60 VESA */
+	{ NULL, 60, 1280, 1024, 9259, 248, 48, 38, 1, 112, 3,  //9259
+	  FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
+	  FB_VMODE_NONINTERLACED, FB_MODE_IS_VESA },
+	  /* 21 1280x1024-75 VESA */
+	{ NULL, 75, 1280, 1024, 7407, 248, 16, 38, 1, 144, 3,
+	  FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
+	  FB_VMODE_NONINTERLACED, FB_MODE_IS_VESA },
+	/* 22 1280x1024-85 VESA */
+	{ NULL, 85, 1280, 1024, 6349, 224, 64, 44, 1, 160, 3,
+	  FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
+	  FB_VMODE_NONINTERLACED, FB_MODE_IS_VESA },
+	
+        /* 34 1024x600-60 VESA */
+        {NULL,  60, 1024, 600, 20423, 144,  40, 18, 1, 104, 3,
+         FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED},
+        /*35 1024x600-70 VESA */
+        {NULL,  70, 1024, 600, 17211, 152,  48, 21, 1, 104, 3,
+         FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED},
+        /* 36 1024x600-75 VESA */
+        {NULL,  75, 1024, 600, 15822, 160,  56, 23, 1, 104, 3,
+         FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED},
+        /* 37 1024x600-85 VESA */
+        {NULL,  85, 1024, 600, 13730, 168,  56, 26, 1, 112, 3,
+         FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED},
+	 /*720x480*/
+        {NULL, 60,  720,  480,  37427, 88,   16, 13, 1,   72,  3,
+        FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED},
+	/* 1368 x 768  Pixel Clock=75hz for lemote*/     
+        {NULL, 60,  1368,  768,  13333, 216, 72, 23, 1,   144,  3,
+        FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED},        
+
+};
+
 
 #define NR_PALETTE	256
+#define PANEL_CRT_ADDR	0x400000
 
 enum sm501_controller {
 	HEAD_CRT	= 0,
 	HEAD_PANEL	= 1,
 };
 
-/* SM501 memory address.
- *
- * This structure is used to track memory usage within the SM501 framebuffer
- * allocation. The sm_addr field is stored as an offset as it is often used
- * against both the physical and mapped addresses.
- */
+/* SM501 memory adress */
 struct sm501_mem {
 	unsigned long	 size;
-	unsigned long	 sm_addr;	/* offset from base of sm501 fb. */
+	unsigned long	 sm_addr;
 	void __iomem	*k_addr;
 };
 
@@ -71,7 +278,6 @@
 	unsigned long		 pm_crt_ctrl;	/* pm: crt ctrl save */
 
 	int			 irq;
-	int			 swap_endian;	/* set to swap rgb=>bgr */
 	void __iomem		*regs;		/* remapped registers */
 	void __iomem		*fbmem;		/* remapped framebuffer */
 	size_t			 fbmem_len;	/* length of remapped region */
@@ -106,6 +312,35 @@
 		var->lower_margin + var->vsync_len;
 }
 
+/*
+ *  Locate in-stock parameter table for the requested mode.
+ *  Success: return a pointer to the mode_parameter_t entry.
+ *  Fail: a NULL pointer.
+ *
+ */
+mode_parameter_t *findVesaModeParam(
+unsigned long width, 
+unsigned long height, 
+unsigned long refresh_rate)
+{
+    mode_parameter_t *mode_table = gModeParamTable;
+
+    /* Walk the entire mode table. */
+    while (mode_table->pixel_clock != 0)
+    {
+        if ((mode_table->horizontal_display_end == width)
+        && (mode_table->vertical_display_end == height)
+        && (mode_table->vertical_frequency == refresh_rate))
+        {
+            return(mode_table);
+        }
+        mode_table++; /* Next entry */
+    }
+
+    /* No match, return NULL pointer */
+    return((mode_parameter_t *)0);
+}
+
 /* sm501fb_sync_regs()
  *
  * This call is mainly for PCI bus systems where we need to
@@ -115,9 +350,10 @@
 
 static inline void sm501fb_sync_regs(struct sm501fb_info *info)
 {
-	readl(info->regs);
+	SmRead32(0);
 }
 
+
 /* sm501_alloc_mem
  *
  * This is an attempt to lay out memory for the two framebuffers and
@@ -139,76 +375,36 @@
  *
  */
 
-#define SM501_MEMF_CURSOR		(1)
-#define SM501_MEMF_PANEL		(2)
-#define SM501_MEMF_CRT			(4)
-#define SM501_MEMF_ACCEL		(8)
-
 static int sm501_alloc_mem(struct sm501fb_info *inf, struct sm501_mem *mem,
 			   unsigned int why, size_t size)
 {
-	struct sm501fb_par *par;
-	struct fb_info *fbi;
-	unsigned int ptr;
-	unsigned int end;
+	unsigned int ptr = 0;
 
 	switch (why) {
 	case SM501_MEMF_CURSOR:
 		ptr = inf->fbmem_len - size;
-		inf->fbmem_len = ptr;	/* adjust available memory. */
+		inf->fbmem_len = ptr;
 		break;
 
 	case SM501_MEMF_PANEL:
+/*
 		ptr = inf->fbmem_len - size;
-		fbi = inf->fb[HEAD_CRT];
-
-		/* round down, some programs such as directfb do not draw
-		 * 0,0 correctly unless the start is aligned to a page start.
-		 */
-
-		if (ptr > 0)
-			ptr &= ~(PAGE_SIZE - 1);
-
-		if (fbi && ptr < fbi->fix.smem_len)
-			return -ENOMEM;
-
-		if (ptr < 0)
+		if (ptr < inf->fb[0]->fix.smem_len)
 			return -ENOMEM;
-
+*/		ptr = 0x0;
 		break;
 
 	case SM501_MEMF_CRT:
-		ptr = 0;
-
-		/* check to see if we have panel memory allocated
-		 * which would put an limit on available memory. */
-
-		fbi = inf->fb[HEAD_PANEL];
-		if (fbi) {
-			par = fbi->par;
-			end = par->screen.k_addr ? par->screen.sm_addr : inf->fbmem_len;
-		} else
-			end = inf->fbmem_len;
-
-		if ((ptr + size) > end)
-			return -ENOMEM;
+		ptr = PANEL_CRT_ADDR;
 
 		break;
 
 	case SM501_MEMF_ACCEL:
-		fbi = inf->fb[HEAD_CRT];
-		ptr = fbi ? fbi->fix.smem_len : 0;
-
-		fbi = inf->fb[HEAD_PANEL];
-		if (fbi) {
-			par = fbi->par;
-			end = par->screen.sm_addr;
-		} else
-			end = inf->fbmem_len;
+		ptr = inf->fb[0]->fix.smem_len;
 
-		if ((ptr + size) > end)
+		if ((ptr + size) >
+		    (inf->fb[1]->fix.smem_start - inf->fbmem_res->start))
 			return -ENOMEM;
-
 		break;
 
 	default:
@@ -219,12 +415,26 @@
 	mem->sm_addr = ptr;
 	mem->k_addr  = inf->fbmem + ptr;
 
-	dev_dbg(inf->dev, "%s: result %08lx, %p - %u, %zd\n",
+	dev_dbg(inf->dev, "%s: result %08lx, %p - %u, %08lx\n",
 		__func__, mem->sm_addr, mem->k_addr, why, size);
 
 	return 0;
 }
 
+
+
+#ifdef CONFIG_MTRR
+
+extern void sm501fb_set_mtrr(struct device *dev);
+
+extern void sm501fb_unset_mtrr(struct device *dev);
+
+#else
+#define sm501fb_set_mtrr(x) do {smdbg("MTRR is disabled in the kernel\n");} while (0)
+
+#define sm501fb_unset_mtrr(x) do { } while (0)
+#endif /* CONFIG_MTRR */
+
 /* sm501fb_ps_to_hz
  *
  * Converts a period in picoseconds to Hz.
@@ -235,7 +445,7 @@
 
 static unsigned long sm501fb_ps_to_hz(unsigned long psvalue)
 {
-	unsigned long long numerator=1000000000000ULL;
+	unsigned long long numerator=1000*1000*1000*1000ULL;
 
 	/* 10^12 / picosecond period gives frequency in Hz */
 	do_div(numerator, psvalue);
@@ -246,6 +456,35 @@
 
 #define sm501fb_hz_to_ps(x) sm501fb_ps_to_hz(x)
 
+
+static void sm501fb_wait_panelvsnc(int vsync_count)
+{
+	unsigned long status;
+
+	while (vsync_count-- > 0)
+	{
+		// Wait for end of vsync.
+		do
+		{
+			status = FIELD_GET(SmRead32(SYS_CTRL),
+							   SYS_CTRL,
+							   PANEL_SYNC);
+		}
+		while (status == SYS_CTRL_PANEL_SYNC_ACTIVE);
+
+		// Wait for start of vsync.
+		do
+		{
+			status = FIELD_GET(SmRead32(SYS_CTRL),
+							   SYS_CTRL, 
+							   PANEL_SYNC);
+		}
+		while (status == SYS_CTRL_PANEL_SYNC_INACTIVE);
+	}
+	smdbg("w_panelvsync\n");
+
+}
+
 /* sm501fb_setup_gamma
  *
  * Programs a linear 1.0 gamma ramp in case the gamma
@@ -260,7 +499,7 @@
 
 	/* set gamma values */
 	for (offset = 0; offset < 256 * 4; offset += 4) {
-		writel(value, fbi->regs + palette + offset);
+		SmWrite32(palette + offset, value);
 		value += 0x010101; 	/* Advance RGB by 1,1,1.*/
 	}
 }
@@ -294,12 +533,16 @@
 
 	if (h_total(var) > 4096 || v_total(var) > 2048)
 		return -EINVAL;
+        /* Commented out by Jason 08/04/2009
+         * to enable vesa_modes that are not 128 bit aligned 
+         * since we've got the workaround with these modes */
 
 	/* check our line length is going to be 128 bit aligned */
 
-	tmp = (var->xres * var->bits_per_pixel) / 8;
+	/* tmp = (var->xres * var->bits_per_pixel) / 8;
 	if ((tmp & 15) != 0)
 		return -EINVAL;
+        */
 
 	/* check the virtual size */
 
@@ -330,40 +573,29 @@
 		break;
 
 	case 16:
-		if (sm->pdata->flags & SM501_FBPD_SWAP_FB_ENDIAN) {
-			var->blue.offset	= 11;
-			var->green.offset	= 5;
-			var->red.offset		= 0;
-		} else {
-			var->red.offset		= 11;
-			var->green.offset	= 5;
-			var->blue.offset	= 0;
-		}
-		var->transp.offset	= 0;
+		var->red.offset		= 11;
+		var->blue.offset	= 0;
+		var->green.offset	= 5;
 
-		var->red.length		= 5;
+		var->red.length	= 5;
 		var->green.length	= 6;
 		var->blue.length	= 5;
+
+		var->transp.offset	= 0;
 		var->transp.length	= 0;
+		
 		break;
 
 	case 32:
-		if (sm->pdata->flags & SM501_FBPD_SWAP_FB_ENDIAN) {
-			var->transp.offset	= 0;
-			var->red.offset		= 8;
-			var->green.offset	= 16;
-			var->blue.offset	= 24;
-		} else {
-			var->transp.offset	= 24;
-			var->red.offset		= 16;
-			var->green.offset	= 8;
-			var->blue.offset	= 0;
-		}
+		var->transp.offset	= 24;
+		var->red.offset		= 16;
+		var->green.offset	= 8;
+		var->blue.offset	= 0;
 
 		var->red.length		= 8;
 		var->green.length	= 8;
 		var->blue.length	= 8;
-		var->transp.length	= 0;
+		var->transp.length	= 8;
 		break;
 
 	default:
@@ -398,6 +630,160 @@
 	return sm501fb_check_var(var, info);
 }
 
+
+static int my_atoi(const char *name)
+{
+    int val = 0;
+
+    for (;; name++) {
+	switch (*name) {
+	    case '0' ... '9':
+		val = 10*val+(*name-'0');
+		break;
+	    default:
+		return val;
+	}
+    }
+}
+
+/* Perform a rounded division. */
+long roundedDiv(long num, long denom)
+{
+    /* n / d + 1 / 2 = (2n + d) / 2d */
+    return (2 * num + denom) / (2 * denom);
+}
+
+
+static int sm501_get_refresh(char * mode)
+{
+	int refresh;
+	char *tmp;
+	if(mode == NULL)
+	{
+		return 60;
+	}
+	tmp = strstr(mode, "@");
+	refresh = my_atoi(tmp + 1);
+	return refresh;
+}
+
+
+/*
+ *  Locate in-stock parameter table for the requested mode.
+ *  Success: return a pointer to the mode_parameter_t entry.
+ *  Fail: a NULL pointer.
+ *
+ */
+mode_parameter_t *find_vesamode(
+unsigned long width, 
+unsigned long height, 
+unsigned long refresh_rate)
+{
+    mode_parameter_t *mode_table = gModeParamTable;
+
+    /* Walk the entire mode table. */
+    while (mode_table->pixel_clock != 0)
+    {
+        if ((mode_table->horizontal_display_end == width)
+        && (mode_table->vertical_display_end == height)
+        && (mode_table->vertical_frequency == refresh_rate))
+        {
+            return(mode_table);
+        }
+        mode_table++; /* Next entry */
+    }
+
+    /* No match, return NULL pointer */
+    return((mode_parameter_t *)0);
+}
+
+/*
+ *  Convert the VESA timing into possible Voyager timing.
+ *  If actual pixel clock is not equal to Vesa timing pixel clock.
+ *  other parameter like horizontal total and sync have to be changed.
+ */
+long sm501fb_adjust_vesamode(struct fb_info *info,
+				     	struct fb_var_screeninfo *var)
+{
+
+    	unsigned long blank_width, sync_start, sync_width, refresh, pixclock;
+	unsigned long horizontal_total, horizontal_sync_start, horizontal_sync_width; 
+	unsigned long reg;
+	mode_parameter_t *pVesaMode; /* physical parameters for the mode */
+	struct sm501fb_par  *par = info->par;
+
+	refresh = sm501_get_refresh(mode_option);
+
+	pVesaMode = find_vesamode(var->xres, var->yres, refresh);
+	if(pVesaMode == NULL)
+	{
+		return -1;
+	}
+	
+	pixclock = sm501fb_ps_to_hz(var->pixclock);
+	if(pixclock == pVesaMode->pixel_clock)
+	{
+		return 0;
+	}
+	
+    	/* Calculate the sync percentages of the VESA mode. */
+	blank_width = pVesaMode->horizontal_total - pVesaMode->horizontal_display_end;
+    	sync_start = roundedDiv((pVesaMode->horizontal_sync_start -
+                       pVesaMode->horizontal_display_end) * 100, blank_width);
+    	sync_width = roundedDiv(pVesaMode->horizontal_sync_width * 100, blank_width);
+
+     	/* Calculate the horizontal total based on the actual pixel clock and VESA line frequency. */
+    	horizontal_total = roundedDiv(pixclock,
+                                    pVesaMode->horizontal_frequency);
+
+    	/* Calculate the sync start and width based on the VESA percentages. */
+    	blank_width = horizontal_total - pVesaMode->horizontal_display_end;
+    	horizontal_sync_start = pVesaMode->horizontal_display_end + roundedDiv(blank_width * sync_start, 100);
+    	horizontal_sync_width = roundedDiv(blank_width * sync_width, 100);
+
+
+	/* program horizontal total */
+	reg = 0;
+	if (par->head == HEAD_CRT)
+	{
+		reg = FIELD_VALUE(reg, CRT_HORIZONTAL_TOTAL, TOTAL, horizontal_total - 1)  | 
+			  FIELD_VALUE(reg, CRT_HORIZONTAL_TOTAL, DISPLAY_END, pVesaMode->horizontal_display_end - 1);
+		SmWrite32(CRT_HORIZONTAL_TOTAL, reg);
+	}
+	#if 0
+	else
+	{
+		reg = FIELD_VALUE(reg, PANEL_HORIZONTAL_TOTAL, TOTAL, horizontal_total - 1) |
+			  FIELD_VALUE(reg, PANEL_HORIZONTAL_TOTAL, TOTAL, pVesaMode->horizontal_display_end - 1);
+		SmWrite32(PANEL_HORIZONTAL_TOTAL, reg);
+	}
+	#endif
+	
+	
+
+	/* program horizontal sync */	
+	reg = 0;
+	if (par->head == HEAD_CRT)	
+	{
+		reg = FIELD_VALUE(reg, CRT_HORIZONTAL_SYNC, WIDTH, horizontal_sync_width) |
+			  FIELD_VALUE(reg, CRT_HORIZONTAL_SYNC, START, horizontal_sync_start - 1);	
+		SmWrite32(CRT_HORIZONTAL_SYNC, reg);
+	}
+	#if 0
+	else
+	{
+		reg = FIELD_VALUE(reg, PANEL_HORIZONTAL_SYNC, WIDTH, horizontal_sync_width) |
+			  FIELD_VALUE(reg, PANEL_HORIZONTAL_SYNC, START, horizontal_sync_start - 1) ;	
+		SmWrite32(PANEL_HORIZONTAL_SYNC, reg);
+	}
+	#endif
+
+	
+	
+    	return 0;
+
+}
+
 /* sm501fb_set_par_common
  *
  * set common registers for framebuffers
@@ -408,8 +794,8 @@
 {
 	struct sm501fb_par  *par = info->par;
 	struct sm501fb_info *fbi = par->info;
-	unsigned long pixclock;      /* pixelclock in Hz */
-	unsigned long sm501pixclock; /* pixelclock the 501 can achive in Hz */
+	unsigned long pixclock=0;      /* pixelclock in Hz */
+	unsigned long sm501pixclock=0; /* pixelclock the 501 can achive in Hz */
 	unsigned int mem_type;
 	unsigned int clock_type;
 	unsigned int head_addr;
@@ -422,13 +808,13 @@
 	case HEAD_CRT:
 		mem_type = SM501_MEMF_CRT;
 		clock_type = SM501_CLOCK_V2XCLK;
-		head_addr = SM501_DC_CRT_FB_ADDR;
+		head_addr = CRT_FB_ADDRESS;
 		break;
 
 	case HEAD_PANEL:
 		mem_type = SM501_MEMF_PANEL;
 		clock_type = SM501_CLOCK_P2XCLK;
-		head_addr = SM501_DC_PANEL_FB_ADDR;
+		head_addr = PANEL_FB_ADDRESS;
 		break;
 
 	default:
@@ -443,46 +829,60 @@
 		break;
 
 	case 16:
-		info->fix.visual = FB_VISUAL_TRUECOLOR;
-		break;
-
 	case 32:
-		info->fix.visual = FB_VISUAL_TRUECOLOR;
+		if (info->var.nonstd)
+			info->fix.visual = FB_VISUAL_DIRECTCOLOR;
+		else
+			info->fix.visual = FB_VISUAL_TRUECOLOR;
+	    	break;
 		break;
 	}
-
 	/* allocate fb memory within 501 */
-	info->fix.line_length = (var->xres_virtual * var->bits_per_pixel)/8;
-	info->fix.smem_len    = info->fix.line_length * var->yres_virtual;
+/*#if 1 // Cheok(7/26/2007):	*/
+        /* Jason(08/05/2009):
+         * to enable the workaround for resolutions that are not 16 pixel aligned
+         */
 
+#if 0 // Cheok(7/26/2007):	
+	info->fix.line_length = (var->xres_virtual * var->bits_per_pixel)/8;
+#else
+	// Cheok(7/26/2007)
+	// Take care cases that screen width is not 16 pixel aligned, which is
+	// a requirement for Direct Draw and the hardware.
+	// Note that standard resolutions: 320, 640., 800, 1024 and 1280 are all
+	// 16 pixel aligned and don't need this adjustment. 
+	// However, new resolution, 1366, has to be adjusted.
+        // fb_width (or Pitch value) has to be 32 byte aligned.
+        {
+                int tmpWidth = var->xres_virtual;
+                tmpWidth = (tmpWidth + 15) & ~15; //This calculation has no effect on 640, 800, 1024 and 1280.
+                info->fix.line_length =( tmpWidth  * var->bits_per_pixel)/8;
+        }
+#endif	
+	info->fix.smem_len  = info->fix.line_length * var->yres_virtual;
 	dev_dbg(fbi->dev, "%s: line length = %u\n", __func__,
 		info->fix.line_length);
-
 	if (sm501_alloc_mem(fbi, &par->screen, mem_type,
 			    info->fix.smem_len)) {
 		dev_err(fbi->dev, "no memory available\n");
 		return -ENOMEM;
 	}
-
 	info->fix.smem_start = fbi->fbmem_res->start + par->screen.sm_addr;
 
 	info->screen_base = fbi->fbmem + par->screen.sm_addr;
 	info->screen_size = info->fix.smem_len;
-
 	/* set start of framebuffer to the screen */
 
-	writel(par->screen.sm_addr | SM501_ADDR_FLIP, fbi->regs + head_addr);
-
+	SmWrite32(PANEL_FB_ADDRESS, 
+					FIELD_VALUE(0,PANEL_FB_ADDRESS, ADDRESS, par->screen.sm_addr) |
+					FIELD_SET(0, PANEL_FB_ADDRESS, STATUS, FLIP));
+	
 	/* program CRT clock  */
-
 	pixclock = sm501fb_ps_to_hz(var->pixclock);
-
-	sm501pixclock = sm501_set_clock(fbi->dev->parent, clock_type,
-					pixclock);
-
+	sm501pixclock = sm501_set_clock(fbi->dev->parent, clock_type, pixclock);
 	/* update fb layer with actual clock used */
 	var->pixclock = sm501fb_hz_to_ps(sm501pixclock);
-
+	
 	dev_dbg(fbi->dev, "%s: pixclock(ps) = %u, pixclock(Hz)  = %lu, "
 	       "sm501pixclock = %lu,  error = %ld%%\n",
 	       __func__, var->pixclock, pixclock, sm501pixclock,
@@ -501,48 +901,63 @@
 {
 	struct sm501fb_par  *par = info->par;
 	struct sm501fb_info *fbi = par->info;
-	void __iomem *base = fbi->regs;
+	unsigned long  base;
 	unsigned long reg;
 
 	if (par->head == HEAD_CRT)
-		base += SM501_DC_CRT_H_TOT;
+		base = CRT_HORIZONTAL_TOTAL;
 	else
-		base += SM501_DC_PANEL_H_TOT;
+		base = PANEL_HORIZONTAL_TOTAL;
 
 	/* set framebuffer width and display width */
 
 	reg = info->fix.line_length;
-	reg |= ((var->xres * var->bits_per_pixel)/8) << 16;
-
-	writel(reg, fbi->regs + (par->head == HEAD_CRT ?
-		    SM501_DC_CRT_FB_OFFSET :  SM501_DC_PANEL_FB_OFFSET));
+	if (par->head == HEAD_CRT)
+		reg = FIELD_VALUE(reg, CRT_FB_WIDTH, WIDTH, (var->xres * var->bits_per_pixel)/8) ;
+	else
+		reg = FIELD_VALUE(reg, PANEL_FB_WIDTH, WIDTH, (var->xres * var->bits_per_pixel)/8) ;
+	SmWrite32((par->head == HEAD_CRT ? CRT_FB_WIDTH :  PANEL_FB_WIDTH), reg);
 
 	/* program horizontal total */
 
-	reg  = (h_total(var) - 1) << 16;
-	reg |= (var->xres - 1);
-
-	writel(reg, base + SM501_OFF_DC_H_TOT);
+	reg = (var->xres - 1);
+	if (par->head == HEAD_CRT)
+		reg = FIELD_VALUE(reg, CRT_HORIZONTAL_TOTAL, TOTAL, (h_total(var) - 1) ) ;
+	else
+		reg = FIELD_VALUE(reg, PANEL_HORIZONTAL_TOTAL, TOTAL, (h_total(var) - 1) ) ;
+	SmWrite32(base+SM501_OFF_DC_H_TOT, reg);
 
 	/* program horizontal sync */
 
-	reg  = var->hsync_len << 16;
-	reg |= var->xres + var->right_margin - 1;
+	reg = var->xres + var->right_margin - 1;
+	if (par->head == HEAD_CRT)
+		reg = FIELD_VALUE(reg, CRT_HORIZONTAL_SYNC, WIDTH, var->hsync_len);
+	else
+		reg = FIELD_VALUE(reg, PANEL_HORIZONTAL_SYNC, WIDTH, var->hsync_len);
 
-	writel(reg, base + SM501_OFF_DC_H_SYNC);
+	SmWrite32(base + SM501_OFF_DC_H_SYNC, reg);
 
 	/* program vertical total */
 
-	reg  = (v_total(var) - 1) << 16;
-	reg |= (var->yres - 1);
-
-	writel(reg, base + SM501_OFF_DC_V_TOT);
+	reg = (var->yres - 1);
+	if (par->head == HEAD_CRT)
+		reg  = FIELD_VALUE(reg, CRT_VERTICAL_TOTAL, TOTAL, (v_total(var) - 1));
+	else
+		reg  = FIELD_VALUE(reg, PANEL_VERTICAL_TOTAL, TOTAL, (v_total(var) - 1));
+	
+	SmWrite32(base + SM501_OFF_DC_V_TOT, reg);
 
 	/* program vertical sync */
-	reg  = var->vsync_len << 16;
-	reg |= var->yres + var->lower_margin - 1;
+	reg  = var->yres + var->lower_margin - 1;
+	if (par->head == HEAD_CRT)
+		reg  = FIELD_VALUE(reg, CRT_VERTICAL_SYNC, HEIGHT, var->vsync_len);
+	else
+		reg  = FIELD_VALUE(reg, PANEL_VERTICAL_SYNC, HEIGHT, var->vsync_len);
+	SmWrite32(base + SM501_OFF_DC_V_SYNC, reg);
+	
+	/*adjust the timing when using dule mode.added by caesar*/
+	sm501fb_adjust_vesamode(info, var);		
 
-	writel(reg, base + SM501_OFF_DC_V_SYNC);
 }
 
 /* sm501fb_pan_crt
@@ -561,15 +976,14 @@
 
 	xoffs = var->xoffset * bytes_pixel;
 
-	reg = readl(fbi->regs + SM501_DC_CRT_CONTROL);
-
-	reg &= ~SM501_DC_CRT_CONTROL_PIXEL_MASK;
-	reg |= ((xoffs & 15) / bytes_pixel) << 4;
-	writel(reg, fbi->regs + SM501_DC_CRT_CONTROL);
-
-	reg = (par->screen.sm_addr + xoffs +
-	       var->yoffset * info->fix.line_length);
-	writel(reg | SM501_ADDR_FLIP, fbi->regs + SM501_DC_CRT_FB_ADDR);
+	reg = SmRead32(CRT_DISPLAY_CTRL);
+	reg = FIELD_VALUE(reg, CRT_DISPLAY_CTRL, PIXEL, (xoffs & 15) / bytes_pixel);
+	
+	SmWrite32(CRT_DISPLAY_CTRL, reg);
+
+	reg = (par->screen.sm_addr + xoffs + var->yoffset * info->fix.line_length);
+	reg = FIELD_SET(reg, CRT_FB_ADDRESS, STATUS , PENDING);
+	SmWrite32(CRT_FB_ADDRESS, reg);
 
 	sm501fb_sync_regs(fbi);
 	return 0;
@@ -587,11 +1001,13 @@
 	struct sm501fb_info *fbi = par->info;
 	unsigned long reg;
 
-	reg = var->xoffset | (var->xres_virtual << 16);
-	writel(reg, fbi->regs + SM501_DC_PANEL_FB_WIDTH);
-
-	reg = var->yoffset | (var->yres_virtual << 16);
-	writel(reg, fbi->regs + SM501_DC_PANEL_FB_HEIGHT);
+	reg = FIELD_VALUE(reg, PANEL_WINDOW_WIDTH, X, var->xoffset);
+	reg = FIELD_VALUE(reg, PANEL_WINDOW_WIDTH, WIDTH, var->xres_virtual);
+	SmWrite32(PANEL_WINDOW_WIDTH, reg);
+
+	reg = FIELD_VALUE(reg, PANEL_WINDOW_HEIGHT, Y, var->yoffset);
+	reg = FIELD_VALUE(reg, PANEL_WINDOW_HEIGHT, HEIGHT, var->yres_virtual);
+	SmWrite32(PANEL_WINDOW_HEIGHT, reg);
 
 	sm501fb_sync_regs(fbi);
 	return 0;
@@ -615,28 +1031,21 @@
 	dev_dbg(fbi->dev, "%s(%p)\n", __func__, info);
 
 	/* enable CRT DAC - note 0 is on!*/
-	sm501_misc_control(fbi->dev->parent, 0, SM501_MISC_DAC_POWER);
-
-	control = readl(fbi->regs + SM501_DC_CRT_CONTROL);
-
-	control &= (SM501_DC_CRT_CONTROL_PIXEL_MASK |
-		    SM501_DC_CRT_CONTROL_GAMMA |
-		    SM501_DC_CRT_CONTROL_BLANK |
-		    SM501_DC_CRT_CONTROL_SEL |
-		    SM501_DC_CRT_CONTROL_CP |
-		    SM501_DC_CRT_CONTROL_TVP);
+	control = SmRead32(MISC_CTRL);
+	SmWrite32(MISC_CTRL, FIELD_SET(control, MISC_CTRL, DAC_POWER, ENABLE) );
+	
+	control = SmRead32(CRT_DISPLAY_CTRL);
 
 	/* set the sync polarities before we check data source  */
 
 	if ((var->sync & FB_SYNC_HOR_HIGH_ACT) == 0)
-		control |= SM501_DC_CRT_CONTROL_HSP;
+		control = FIELD_SET(control, CRT_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_LOW);
 
 	if ((var->sync & FB_SYNC_VERT_HIGH_ACT) == 0)
-		control |= SM501_DC_CRT_CONTROL_VSP;
+		control = FIELD_SET(control, CRT_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_LOW);
 
-	if ((control & SM501_DC_CRT_CONTROL_SEL) == 0) {
+	if ( FIELD_GET(control, CRT_DISPLAY_CTRL, SELECT)== CRT_DISPLAY_CTRL_SELECT_PANEL) {
 		/* the head is displaying panel data... */
-
 		sm501_alloc_mem(fbi, &par->screen, SM501_MEMF_CRT, 0);
 		goto out_update;
 	}
@@ -650,122 +1059,117 @@
 	sm501fb_pan_crt(var, info);
 	sm501fb_set_par_geometry(info, var);
 
-	control |= SM501_FIFO_3;	/* fill if >3 free slots */
-
+	dev_dbg(fbi->dev, "CRT pixel:%d\n",var->bits_per_pixel);
+	control = FIELD_SET(control, CRT_DISPLAY_CTRL, FIFO, 3);
+	
 	switch(var->bits_per_pixel) {
 	case 8:
-		control |= SM501_DC_CRT_CONTROL_8BPP;
+		control = FIELD_SET(control, CRT_DISPLAY_CTRL, FORMAT,  8);
 		break;
-
 	case 16:
-		control |= SM501_DC_CRT_CONTROL_16BPP;
-		sm501fb_setup_gamma(fbi, SM501_DC_CRT_PALETTE);
+		control = FIELD_SET(control, CRT_DISPLAY_CTRL, FORMAT,  16);
+		sm501fb_setup_gamma(fbi, CRT_PALETTE_RAM);
 		break;
-
 	case 32:
-		control |= SM501_DC_CRT_CONTROL_32BPP;
-		sm501fb_setup_gamma(fbi, SM501_DC_CRT_PALETTE);
+		control = FIELD_SET(control, CRT_DISPLAY_CTRL, FORMAT,  32);
+		sm501fb_setup_gamma(fbi, CRT_PALETTE_RAM);
 		break;
-
 	default:
 		BUG();
 	}
 
-	control |= SM501_DC_CRT_CONTROL_SEL;	/* CRT displays CRT data */
-	control |= SM501_DC_CRT_CONTROL_TE;	/* enable CRT timing */
-	control |= SM501_DC_CRT_CONTROL_ENABLE;	/* enable CRT plane */
-
+#ifdef CONFIG_FB_SM501_DUAL_HEAD
+	control = FIELD_SET(control, CRT_DISPLAY_CTRL, SELECT,  CRT);
+#else
+	control = FIELD_SET(control, CRT_DISPLAY_CTRL, SELECT,  PANEL);
+#endif
+	control = FIELD_SET(control, CRT_DISPLAY_CTRL, TIMING,  ENABLE);
+	control = FIELD_SET(control, CRT_DISPLAY_CTRL, PLANE, ENABLE);
+	
  out_update:
 	dev_dbg(fbi->dev, "new control is %08lx\n", control);
 
-	writel(control, fbi->regs + SM501_DC_CRT_CONTROL);
+	SmWrite32(CRT_DISPLAY_CTRL, control);
 	sm501fb_sync_regs(fbi);
 
 	return 0;
 }
 
-static void sm501fb_panel_power(struct sm501fb_info *fbi, int to)
+
+/* sm501fb_panel_power
+ *
+ * Set power of panel
+*/
+static void sm501fb_panel_power(panel_state_t on_off, int vsync_delay)
 {
-	unsigned long control;
-	void __iomem *ctrl_reg = fbi->regs + SM501_DC_PANEL_CONTROL;
-	struct sm501_platdata_fbsub *pd = fbi->pdata->fb_pnl;
+	unsigned long panelControl = SmRead32(PANEL_DISPLAY_CTRL);
 
-	control = readl(ctrl_reg);
+	if (on_off == PANEL_ON)
+	{
+		// Turn on FPVDDEN.
+		panelControl = FIELD_SET(panelControl,
+								 PANEL_DISPLAY_CTRL, FPVDDEN, HIGH);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
 
-	if (to && (control & SM501_DC_PANEL_CONTROL_VDD) == 0) {
-		/* enable panel power */
+		sm501fb_wait_panelvsnc(vsync_delay);
 
-		control |= SM501_DC_PANEL_CONTROL_VDD;	/* FPVDDEN */
-		writel(control, ctrl_reg);
-		sm501fb_sync_regs(fbi);
-		mdelay(10);
+		// Turn on FPDATA.
+		panelControl = FIELD_SET(panelControl, 
+								 PANEL_DISPLAY_CTRL, DATA, ENABLE);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
 
-		control |= SM501_DC_PANEL_CONTROL_DATA;	/* DATA */
-		writel(control, ctrl_reg);
-		sm501fb_sync_regs(fbi);
-		mdelay(10);
+		sm501fb_wait_panelvsnc(vsync_delay);
 
-		/* VBIASEN */
+		// Turn on FPVBIAS.
+		panelControl = FIELD_SET(panelControl, 
+								 PANEL_DISPLAY_CTRL, VBIASEN, HIGH);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
 
-		if (!(pd->flags & SM501FB_FLAG_PANEL_NO_VBIASEN)) {
-			if (pd->flags & SM501FB_FLAG_PANEL_INV_VBIASEN)
-				control &= ~SM501_DC_PANEL_CONTROL_BIAS;
-			else
-				control |= SM501_DC_PANEL_CONTROL_BIAS;
+		sm501fb_wait_panelvsnc(vsync_delay);
 
-			writel(control, ctrl_reg);
-			sm501fb_sync_regs(fbi);
-			mdelay(10);
-		}
+		// Turn on FPEN.
+		panelControl = FIELD_SET(panelControl, 
+								 PANEL_DISPLAY_CTRL, FPEN, HIGH);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
+	}
 
-		if (!(pd->flags & SM501FB_FLAG_PANEL_NO_FPEN)) {
-			if (pd->flags & SM501FB_FLAG_PANEL_INV_FPEN)
-				control &= ~SM501_DC_PANEL_CONTROL_FPEN;
-			else
-				control |= SM501_DC_PANEL_CONTROL_FPEN;
+	else
+	{
+		// Turn off FPEN.
+		panelControl = FIELD_SET(panelControl,
+								 PANEL_DISPLAY_CTRL, FPEN, LOW);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
 
-			writel(control, ctrl_reg);
-			sm501fb_sync_regs(fbi);
-			mdelay(10);
-		}
-	} else if (!to && (control & SM501_DC_PANEL_CONTROL_VDD) != 0) {
-		/* disable panel power */
-		if (!(pd->flags & SM501FB_FLAG_PANEL_NO_FPEN)) {
-			if (pd->flags & SM501FB_FLAG_PANEL_INV_FPEN)
-				control |= SM501_DC_PANEL_CONTROL_FPEN;
-			else
-				control &= ~SM501_DC_PANEL_CONTROL_FPEN;
+		sm501fb_wait_panelvsnc(vsync_delay);
 
-			writel(control, ctrl_reg);
-			sm501fb_sync_regs(fbi);
-			mdelay(10);
-		}
 
-		if (!(pd->flags & SM501FB_FLAG_PANEL_NO_VBIASEN)) {
-			if (pd->flags & SM501FB_FLAG_PANEL_INV_VBIASEN)
-				control |= SM501_DC_PANEL_CONTROL_BIAS;
-			else
-				control &= ~SM501_DC_PANEL_CONTROL_BIAS;
+		// Turn off FPVBIASEN.
+		panelControl = FIELD_SET(panelControl, 
+								 PANEL_DISPLAY_CTRL, VBIASEN, LOW);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
 
-			writel(control, ctrl_reg);
-			sm501fb_sync_regs(fbi);
-			mdelay(10);
-		}
+		sm501fb_wait_panelvsnc(vsync_delay);
+
+
+		// Turn off FPDATA.
+		panelControl = FIELD_SET(panelControl, 
+								 PANEL_DISPLAY_CTRL, DATA, DISABLE);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
+
+		sm501fb_wait_panelvsnc(vsync_delay);
 
-		control &= ~SM501_DC_PANEL_CONTROL_DATA;
-		writel(control, ctrl_reg);
-		sm501fb_sync_regs(fbi);
-		mdelay(10);
-
-		control &= ~SM501_DC_PANEL_CONTROL_VDD;
-		writel(control, ctrl_reg);
-		sm501fb_sync_regs(fbi);
-		mdelay(10);
+
+		// Turn off FPVDDEN.
+		panelControl = FIELD_SET(panelControl, 
+								 PANEL_DISPLAY_CTRL, FPVDDEN, LOW);
+		SmWrite32(PANEL_DISPLAY_CTRL, panelControl);
 	}
+	
 
-	sm501fb_sync_regs(fbi);
 }
 
+
+
 /* sm501fb_set_par_pnl
  *
  * Set the panel video mode from the fb_info structure
@@ -793,72 +1197,60 @@
 
 	/* update control register */
 
-	control = readl(fbi->regs + SM501_DC_PANEL_CONTROL);
-	control &= (SM501_DC_PANEL_CONTROL_GAMMA |
-		    SM501_DC_PANEL_CONTROL_VDD  |
-		    SM501_DC_PANEL_CONTROL_DATA |
-		    SM501_DC_PANEL_CONTROL_BIAS |
-		    SM501_DC_PANEL_CONTROL_FPEN |
-		    SM501_DC_PANEL_CONTROL_CP |
-		    SM501_DC_PANEL_CONTROL_CK |
-		    SM501_DC_PANEL_CONTROL_HP |
-		    SM501_DC_PANEL_CONTROL_VP |
-		    SM501_DC_PANEL_CONTROL_HPD |
-		    SM501_DC_PANEL_CONTROL_VPD);
+	control = SmRead32(PANEL_DISPLAY_CTRL);
 
-	control |= SM501_FIFO_3;	/* fill if >3 free slots */
+	control = FIELD_SET(control, PANEL_DISPLAY_CTRL, FIFO, 3);	
+
+	dev_dbg(fbi->dev, "PANEL pixel:%d\n",var->bits_per_pixel);
 
 	switch(var->bits_per_pixel) {
 	case 8:
-		control |= SM501_DC_PANEL_CONTROL_8BPP;
+		control = FIELD_SET(control, PANEL_DISPLAY_CTRL, FORMAT,  8);
 		break;
-
 	case 16:
-		control |= SM501_DC_PANEL_CONTROL_16BPP;
-		sm501fb_setup_gamma(fbi, SM501_DC_PANEL_PALETTE);
+		control = FIELD_SET(control, PANEL_DISPLAY_CTRL, FORMAT,  16);
+		sm501fb_setup_gamma(fbi, PANEL_PALETTE_RAM);
 		break;
-
 	case 32:
-		control |= SM501_DC_PANEL_CONTROL_32BPP;
-		sm501fb_setup_gamma(fbi, SM501_DC_PANEL_PALETTE);
+		control = FIELD_SET(control, PANEL_DISPLAY_CTRL, FORMAT,  32);
+		sm501fb_setup_gamma(fbi, PANEL_PALETTE_RAM);
 		break;
-
 	default:
 		BUG();
 	}
 
-	writel(0x0, fbi->regs + SM501_DC_PANEL_PANNING_CONTROL);
+	SmWrite32(PANEL_PAN_CTRL, 0x0);
 
 	/* panel plane top left and bottom right location */
 
-	writel(0x00, fbi->regs + SM501_DC_PANEL_TL_LOC);
+	SmWrite32(PANEL_PLANE_TL, 0x0);
 
 	reg  = var->xres - 1;
 	reg |= (var->yres - 1) << 16;
-
-	writel(reg, fbi->regs + SM501_DC_PANEL_BR_LOC);
+	SmWrite32( PANEL_PLANE_BR, reg);
 
 	/* program panel control register */
 
-	control |= SM501_DC_PANEL_CONTROL_TE;	/* enable PANEL timing */
-	control |= SM501_DC_PANEL_CONTROL_EN;	/* enable PANEL gfx plane */
+	control = FIELD_SET(control, PANEL_DISPLAY_CTRL, TIMING, ENABLE);	/* DATA */
+	control = FIELD_SET(control, PANEL_DISPLAY_CTRL, PLANE, ENABLE);	/* DATA */
 
 	if ((var->sync & FB_SYNC_HOR_HIGH_ACT) == 0)
-		control |= SM501_DC_PANEL_CONTROL_HSP;
+		control = FIELD_SET(control, PANEL_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_LOW);	/* DATA */
 
 	if ((var->sync & FB_SYNC_VERT_HIGH_ACT) == 0)
-		control |= SM501_DC_PANEL_CONTROL_VSP;
-
-	writel(control, fbi->regs + SM501_DC_PANEL_CONTROL);
-	sm501fb_sync_regs(fbi);
+		control = FIELD_SET(control, PANEL_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_LOW);	/* DATA */
 
-	/* ensure the panel interface is not tristated at this point */
+	dev_dbg(fbi->dev, "%s H:%d  V:%d\n", __func__, var->sync, var->sync);
 
-	sm501_modify_reg(fbi->dev->parent, SM501_SYSTEM_CONTROL,
-			 0, SM501_SYSCTRL_PANEL_TRISTATE);
+	SmWrite32(PANEL_DISPLAY_CTRL, control);
+	sm501fb_wait_panelvsnc(4);
 
 	/* power the panel up */
-	sm501fb_panel_power(fbi, 1);
+
+	sm501fb_panel_power(PANEL_ON, 4);
+
+	deInit(info->var.xres, info->var.yres, info->var.bits_per_pixel);
+
 	return 0;
 }
 
@@ -889,40 +1281,63 @@
 {
 	struct sm501fb_par  *par = info->par;
 	struct sm501fb_info *fbi = par->info;
-	void __iomem *base = fbi->regs;
+	unsigned long base;
 	unsigned int val;
+	u32 *pal = par->pseudo_palette;
 
+	if (regno > 255)
+		return 1;
+	
 	if (par->head == HEAD_CRT)
-		base += SM501_DC_CRT_PALETTE;
+		base = CRT_PALETTE_RAM;
 	else
-		base += SM501_DC_PANEL_PALETTE;
+		base = PANEL_PALETTE_RAM;
+
+	if(info->var.grayscale)
+		red = green = blue = (red * 77 + green * 151 + blue * 28) >> 8;
 
 	switch (info->fix.visual) {
+	case FB_VISUAL_DIRECTCOLOR:
 	case FB_VISUAL_TRUECOLOR:
-		/* true-colour, use pseuo-palette */
-
-		if (regno < 16) {
-			u32 *pal = par->pseudo_palette;
-
-			val  = chan_to_field(red,   &info->var.red);
-			val |= chan_to_field(green, &info->var.green);
-			val |= chan_to_field(blue,  &info->var.blue);
-
-			pal[regno] = val;
+		/* 16/32 bit true-colour, use pseuo-palette for 16 base color*/
+		if (regno < 16) 
+		{			
+			if (info->var.bits_per_pixel==16)
+			{				
+				val  = chan_to_field(red,   &info->var.red);
+				val |= chan_to_field(green, &info->var.green);
+				val |= chan_to_field(blue,  &info->var.blue);
+#ifdef __BIG_ENDIAN
+				 pal[regno] =( (red & 0xf800) >> 8) |
+				 			((green & 0xe000) >> 13) |
+				 			((green & 0x1c00) << 3) |
+				 			((blue & 0xf800) >> 3);
+#else
+				pal[regno] = val;
+#endif
+			}
+			else
+			{				
+				val  = chan_to_field(red,   &info->var.red);
+				val |= chan_to_field(green, &info->var.green);
+				val |= chan_to_field(blue,  &info->var.blue);
+#ifdef __BIG_ENDIAN				
+				pal[regno] = ((val & 0x00ff)<<24)|
+							((val & 0xff00)<<8)|
+							((val & 0xff0000)>>8);				
+#else
+				pal[regno] = val;
+#endif
+			}
 		}
 		break;
-
 	case FB_VISUAL_PSEUDOCOLOR:
-		if (regno < 256) {
-			val = (red >> 8) << 16;
-			val |= (green >> 8) << 8;
-			val |= blue >> 8;
-
-			writel(val, base + (regno * 4));
-		}
-
+		/* color depth 8 bit*/
+			val = ((red >> 8) <<16 )|
+				((green >> 8) <<8)|
+				((blue >> 8));		
+			SmWrite32(base + (regno * 4), val);
 		break;
-
 	default:
 		return 1;   /* unknown type */
 	}
@@ -930,76 +1345,116 @@
 	return 0;
 }
 
-/* sm501fb_blank_pnl
+
+/* field_to_chan
  *
- * Blank or un-blank the panel interface
+ * convert  a field position intoa colour value
+ *
+ * from pxafb.c
 */
 
-static int sm501fb_blank_pnl(int blank_mode, struct fb_info *info)
+static inline u16 field_to_chan(unsigned int chan,
+					 struct fb_bitfield *bf)
 {
-	struct sm501fb_par  *par = info->par;
-	struct sm501fb_info *fbi = par->info;
+	chan >>= bf->offset;
+	chan &=( (1<<bf->length)-1);
+	chan <<= 16 - bf->length;	
+	return chan &0xffff;
+}
 
-	dev_dbg(fbi->dev, "%s(mode=%d, %p)\n", __func__, blank_mode, info);
+int sm501fb_set_cmap(struct fb_cmap *cmap, struct fb_info *info)
+{
+	int i, start, rc = 0;
+	u16 *red, *green, *blue, *transp;
+	u_int hred, hgreen, hblue, htransp= 0xffff;
+	unsigned int val;
 
-	switch (blank_mode) {
-	case FB_BLANK_POWERDOWN:
-		sm501fb_panel_power(fbi, 0);
-		break;
+	red = cmap->red;
+	green = cmap->green;
+	blue = cmap->blue;
+	transp = cmap->transp;
+	start = cmap->start;
 
-	case FB_BLANK_UNBLANK:
-		sm501fb_panel_power(fbi, 1);
-		break;
+	if (start < 0 || (!info->fbops->fb_setcolreg))
+		return -EINVAL;
 
-	case FB_BLANK_NORMAL:
-	case FB_BLANK_VSYNC_SUSPEND:
-	case FB_BLANK_HSYNC_SUSPEND:
-	default:
-		return 1;
-	}
+		for (i = 0; i < cmap->len; i++) {
+#if 1
+			val  = chan_to_field(red[i],   &info->var.red);
+			val |= chan_to_field(green[i], &info->var.green);
+			val |= chan_to_field(blue[i],  &info->var.blue);
+			
+			if (transp){
+				val |= chan_to_field(transp[i],  &info->var.transp);
+			}
 
-	return 0;
+			val = (val& 0xff00ff00>>8)|(val& 0x00ff00ff<<8);
+			
+			red[i] |= field_to_chan(val,&info->var.red);
+			green[i] |= field_to_chan(val,&info->var.green);
+			blue[i] |= field_to_chan(val,&info->var.blue);
+
+			if (transp)
+				transp[i] |= field_to_chan(val,&info->var.transp);
+#else
+
+		           red[i] = i*65535/((1<<info->var.red.length)-1);
+	 	           green[i] = i*65535/((1<<info->var.green.length)-1);
+	 	           blue[i] = i*65535/((1<<info->var.blue.length)-1);
+	 	           transp[i] = 0;
+#endif			   
+			if (info->fbops->fb_setcolreg(start++,
+						      red[i], green[i], blue[i],
+						      transp[i], info))
+			break;
+		}
+		
+	if (rc == 0)
+		fb_copy_cmap(cmap, &info->cmap);
+
+	return rc;
 }
 
-/* sm501fb_blank_crt
+/* sm501fb_blank
  *
  * Blank or un-blank the crt interface
 */
 
-static int sm501fb_blank_crt(int blank_mode, struct fb_info *info)
+static int sm501fb_blank(int blank_mode, struct fb_info *info)
 {
 	struct sm501fb_par  *par = info->par;
 	struct sm501fb_info *fbi = par->info;
-	unsigned long ctrl;
+	unsigned long ctrl, misc_value;
 
 	dev_dbg(fbi->dev, "%s(mode=%d, %p)\n", __func__, blank_mode, info);
 
-	ctrl = readl(fbi->regs + SM501_DC_CRT_CONTROL);
+	ctrl = SmRead32(SYSTEM_CTRL);
 
 	switch (blank_mode) {
 	case FB_BLANK_POWERDOWN:
-		ctrl &= ~SM501_DC_CRT_CONTROL_ENABLE;
-		sm501_misc_control(fbi->dev->parent, SM501_MISC_DAC_POWER, 0);
-
+		ctrl = FIELD_SET(ctrl, SYSTEM_CTRL, DPMS, VNHN);
+		
 	case FB_BLANK_NORMAL:
-		ctrl |= SM501_DC_CRT_CONTROL_BLANK;
+		ctrl = FIELD_SET(ctrl, SYSTEM_CTRL, DPMS, VPHP);
 		break;
 
 	case FB_BLANK_UNBLANK:
-		ctrl &= ~SM501_DC_CRT_CONTROL_BLANK;
-		ctrl |=  SM501_DC_CRT_CONTROL_ENABLE;
-		sm501_misc_control(fbi->dev->parent, 0, SM501_MISC_DAC_POWER);
+		ctrl = FIELD_SET(ctrl, SYSTEM_CTRL, DPMS, VPHP);
 		break;
 
 	case FB_BLANK_VSYNC_SUSPEND:
+		ctrl = FIELD_SET(ctrl, SYSTEM_CTRL, DPMS, VNHP);
+		break;		
 	case FB_BLANK_HSYNC_SUSPEND:
+		ctrl = FIELD_SET(ctrl, SYSTEM_CTRL, DPMS, VPHN);
+		break;
 	default:
 		return 1;
 
 	}
 
-	writel(ctrl, fbi->regs + SM501_DC_CRT_CONTROL);
-	sm501fb_sync_regs(fbi);
+	SmWrite32(SYSTEM_CTRL, ctrl);
+	sm501fb_wait_panelvsnc(4);
 
 	return 0;
 }
@@ -1013,16 +1468,16 @@
 {
 	struct sm501fb_par  *par = info->par;
 	struct sm501fb_info *fbi = par->info;
-	void __iomem *base = fbi->regs;
+	unsigned long base;
 	unsigned long hwc_addr;
 	unsigned long fg, bg;
 
-	dev_dbg(fbi->dev, "%s(%p,%p)\n", __func__, info, cursor);
+//	dev_dbg(fbi->dev, "%s(%p,%p)\n", __func__, info, cursor);
 
 	if (par->head == HEAD_CRT)
-		base += SM501_DC_CRT_HWC_BASE;
+		base = CRT_HWC_ADDRESS;
 	else
-		base += SM501_DC_PANEL_HWC_BASE;
+		base = PANEL_HWC_ADDRESS;
 
 	/* check not being asked to exceed capabilities */
 
@@ -1035,12 +1490,12 @@
 	if (cursor->image.depth > 1)
 		return -EINVAL;
 
-	hwc_addr = readl(base + SM501_OFF_HWC_ADDR);
+	hwc_addr = SmRead32(base + SM501_OFF_HWC_ADDR);
 
 	if (cursor->enable)
-		writel(hwc_addr | SM501_HWC_EN, base + SM501_OFF_HWC_ADDR);
+		SmWrite32( base + SM501_OFF_HWC_ADDR, hwc_addr | SM501_HWC_EN);
 	else
-		writel(hwc_addr & ~SM501_HWC_EN, base + SM501_OFF_HWC_ADDR);
+		SmWrite32(base + SM501_OFF_HWC_ADDR, hwc_addr & ~SM501_HWC_EN );
 
 	/* set data */
 	if (cursor->set & FB_CUR_SETPOS) {
@@ -1050,11 +1505,11 @@
 		if (x >= 2048 || y >= 2048 )
 			return -EINVAL;
 
-		dev_dbg(fbi->dev, "set position %d,%d\n", x, y);
+//		dev_dbg(fbi->dev, "set position %d,%d\n", x, y);
 
 		//y += cursor->image.height;
 
-		writel(x | (y << 16), base + SM501_OFF_HWC_LOC);
+		SmWrite32(base + SM501_OFF_HWC_LOC, x | (y << 16));
 	}
 
 	if (cursor->set & FB_CUR_SETCMAP) {
@@ -1072,10 +1527,10 @@
 			((info->cmap.green[fg_col] & 0xFC) << 3) |
 			((info->cmap.blue[fg_col] & 0xF8) >> 3);
 
-		dev_dbg(fbi->dev, "fgcol %08lx, bgcol %08lx\n", fg, bg);
+		dev_dbg(fbi->dev, "fgcol %08x, bgcol %08x\n", fg, bg);
 
-		writel(bg, base + SM501_OFF_HWC_COLOR_1_2);
-		writel(fg, base + SM501_OFF_HWC_COLOR_3);
+		SmWrite32(base + SM501_OFF_HWC_COLOR_1_2, bg);
+		SmWrite32(base + SM501_OFF_HWC_COLOR_3, fg);
 	}
 
 	if (cursor->set & FB_CUR_SETSIZE ||
@@ -1092,15 +1547,16 @@
 		unsigned char  dmsk = 0;
 		unsigned int   op;
 
-		dev_dbg(fbi->dev, "%s: setting shape (%d,%d)\n",
-			__func__, cursor->image.width, cursor->image.height);
+//		dev_dbg(fbi->dev, "%s: setting shape (%d,%d)\n", __func__, cursor->image.width, cursor->image.height);
 
 		for (op = 0; op < (64*64*2)/8; op+=4)
 			writel(0x0, dst + op);
 
 		for (y = 0; y < cursor->image.height; y++) {
-			for (x = 0; x < cursor->image.width; x++) {
-				if ((x % 8) == 0) {
+			for (x = 0; x < cursor->image.width; x++) 
+			{
+				if ((x % 8) == 0) 
+				{
 					dcol = *pcol++;
 					dmsk = *pmsk++;
 				} else {
@@ -1108,7 +1564,8 @@
 					dmsk >>= 1;
 				}
 
-				if (dmsk & 1) {
+				if (dmsk & 1) 
+				{
 					op = (dcol & 1) ? 1 : 3;
 					op <<= ((x % 4) * 2);
 
@@ -1124,10 +1581,182 @@
 	return 0;
 }
 
-/* sm501fb_crtsrc_show
- *
- * device attribute code to show where the crt output is sourced from
-*/
+
+static ssize_t
+sm501fb_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
+{
+	unsigned long p = *ppos;
+	struct inode *inode = file->f_path.dentry->d_inode;
+	int fbidx = iminor(inode);
+	struct fb_info *info = registered_fb[fbidx];
+	u32 *buffer, *dst;
+	u32 __iomem *src;
+	int c, i, cnt = 0, err = 0;
+	unsigned long total_size;
+
+	if (!info || ! info->screen_base)
+		return -ENODEV;
+
+	if (info->state != FBINFO_STATE_RUNNING)
+		return -EPERM;
+
+	total_size = info->screen_size;
+
+	if (total_size == 0)
+		total_size = info->fix.smem_len;
+
+	if (p >= total_size)
+		return 0;
+
+	if (count >= total_size)
+		count = total_size;
+
+	if (count + p > total_size)
+		count = total_size - p;
+
+	buffer = kmalloc((count > PAGE_SIZE) ? PAGE_SIZE : count,
+			 GFP_KERNEL);
+	if (!buffer)
+		return -ENOMEM;
+
+	src = (u32 __iomem *) (info->screen_base + p);
+
+	if (info->fbops->fb_sync)
+		info->fbops->fb_sync(info);
+
+	while (count) {
+		c  = (count > PAGE_SIZE) ? PAGE_SIZE : count;
+		dst = buffer;
+		for (i = c >> 2; i--; ){
+			*dst = fb_readl(src++);
+			*dst  = (*dst & 0xff00ff00>>8)|(*dst & 0x00ff00ff<<8);
+			dst++;
+			}
+		if (c & 3) {
+			u8 *dst8 = (u8 *) dst;
+			u8 __iomem *src8 = (u8 __iomem *) src;
+
+			for (i = c & 3; i--;){
+				if (i&1){
+				*dst8++ = fb_readb(++src8);
+				}
+				else{
+				*dst8++ = fb_readb(--src8);
+				src8 +=2;
+				}
+			}
+			src = (u32 __iomem *) src8;
+		}
+
+		if (copy_to_user(buf, buffer, c)) {
+			err = -EFAULT;
+			break;
+		}
+		*ppos += c;
+		buf += c;
+		cnt += c;
+		count -= c;
+	}
+
+	kfree(buffer);
+
+	return (err) ? err : cnt;
+}
+
+static ssize_t
+sm501fb_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
+{
+	unsigned long p = *ppos;
+	struct inode *inode = file->f_path.dentry->d_inode;
+	int fbidx = iminor(inode);
+	struct fb_info *info = registered_fb[fbidx];
+	u32 *buffer, *src;
+	u32 __iomem *dst;
+	int c, i, cnt = 0, err = 0;
+	unsigned long total_size;
+
+	if (!info || !info->screen_base)
+		return -ENODEV;
+
+	if (info->state != FBINFO_STATE_RUNNING)
+		return -EPERM;
+
+	
+	total_size = info->screen_size;
+
+	if (total_size == 0)
+		total_size = info->fix.smem_len;
+
+	if (p > total_size)
+		return -EFBIG;
+
+	if (count > total_size) {
+		err = -EFBIG;
+		count = total_size;
+	}
+
+	if (count + p > total_size) {
+		if (!err)
+			err = -ENOSPC;
+
+		count = total_size - p;
+	}
+
+	buffer = kmalloc((count > PAGE_SIZE) ? PAGE_SIZE : count,
+			 GFP_KERNEL);
+	if (!buffer)
+		return -ENOMEM;
+
+	dst = (u32 __iomem *) (info->screen_base + p);
+
+	if (info->fbops->fb_sync)
+		info->fbops->fb_sync(info);
+
+	while (count) {
+		c = (count > PAGE_SIZE) ? PAGE_SIZE : count;
+		src = buffer;
+
+		if (copy_from_user(src, buf, c)) {
+			err = -EFAULT;
+			break;
+		}
+
+		for (i = c >> 2; i--; ){
+			fb_writel( (*src& 0xff00ff00>>8)|(*src& 0x00ff00ff<<8), dst++);
+			src++;
+		}
+		if (c & 3) {
+			u8 *src8 = (u8 *) src;
+			u8 __iomem *dst8 = (u8 __iomem *) dst;
+
+			for (i = c & 3; i--; ){
+				if (i&1){
+				fb_writeb(*src8++, ++dst8);
+				}
+				else{
+				fb_writeb(*src8++, --dst8);
+				dst8 +=2;
+				}
+			}
+			dst = (u32 __iomem *) dst8;
+		}
+
+		*ppos += c;
+		buf += c;
+		cnt += c;
+		count -= c;
+	}
+
+	kfree(buffer);
+
+	return (cnt) ? cnt : err;
+}
+
+
+/* sm501fb_crtsrc_show
+ *
+ * device attribute code to show where the crt output is sourced from
+*/
 
 static ssize_t sm501fb_crtsrc_show(struct device *dev,
 			       struct device_attribute *attr, char *buf)
@@ -1135,8 +1764,8 @@
 	struct sm501fb_info *info = dev_get_drvdata(dev);
 	unsigned long ctrl;
 
-	ctrl = readl(info->regs + SM501_DC_CRT_CONTROL);
-	ctrl &= SM501_DC_CRT_CONTROL_SEL;
+	ctrl = SmRead32(CRT_DISPLAY_CTRL);
+	ctrl = FIELD_GET(ctrl, CRT_DISPLAY_CTRL, SELECT);
 
 	return snprintf(buf, PAGE_SIZE, "%s\n", ctrl ? "crt" : "panel");
 }
@@ -1166,19 +1795,24 @@
 
 	dev_info(dev, "setting crt source to head %d\n", head);
 
-	ctrl = readl(info->regs + SM501_DC_CRT_CONTROL);
+	ctrl = SmRead32(CRT_DISPLAY_CTRL);
 
 	if (head == HEAD_CRT) {
-		ctrl |= SM501_DC_CRT_CONTROL_SEL;
-		ctrl |= SM501_DC_CRT_CONTROL_ENABLE;
-		ctrl |= SM501_DC_CRT_CONTROL_TE;
+
+#ifdef CONFIG_FB_SM501_DUAL_HEAD
+		ctrl = FIELD_SET(ctrl, CRT_DISPLAY_CTRL, SELECT, CRT)||
+#else
+		ctrl = FIELD_SET(ctrl, CRT_DISPLAY_CTRL, SELECT, PANEL)||
+#endif
+			FIELD_SET(ctrl, CRT_DISPLAY_CTRL, PLANE, ENABLE)||
+				FIELD_SET(ctrl, CRT_DISPLAY_CTRL, TIMING, ENABLE);
 	} else {
-		ctrl &= ~SM501_DC_CRT_CONTROL_SEL;
-		ctrl &= ~SM501_DC_CRT_CONTROL_ENABLE;
-		ctrl &= ~SM501_DC_CRT_CONTROL_TE;
+		ctrl = FIELD_SET(ctrl, CRT_DISPLAY_CTRL, SELECT, PANEL)||
+			FIELD_SET(ctrl, CRT_DISPLAY_CTRL, PLANE, DISABLE)||
+				FIELD_SET(ctrl, CRT_DISPLAY_CTRL, TIMING, DISABLE);
 	}
-
-	writel(ctrl, info->regs + SM501_DC_CRT_CONTROL);
+	SmWrite32(CRT_DISPLAY_CTRL, ctrl);
+	
 	sm501fb_sync_regs(info);
 
 	return len;
@@ -1194,12 +1828,11 @@
 static int sm501fb_show_regs(struct sm501fb_info *info, char *ptr,
 			     unsigned int start, unsigned int len)
 {
-	void __iomem *mem = info->regs;
 	char *buf = ptr;
 	unsigned int reg;
 
 	for (reg = start; reg < (len + start); reg += 4)
-		ptr += sprintf(ptr, "%08x = %08x\n", reg, readl(mem + reg));
+		ptr += sprintf(ptr, "%08x = %08x\n", reg, SmRead32(reg));
 
 	return ptr - buf;
 }
@@ -1215,8 +1848,8 @@
 	struct sm501fb_info *info = dev_get_drvdata(dev);
 	char *ptr = buf;
 
-	ptr += sm501fb_show_regs(info, ptr, SM501_DC_CRT_CONTROL, 0x40);
-	ptr += sm501fb_show_regs(info, ptr, SM501_DC_CRT_HWC_BASE, 0x10);
+	ptr += sm501fb_show_regs(info, ptr, CRT_DISPLAY_CTRL, 0x40);
+	ptr += sm501fb_show_regs(info, ptr, CRT_HWC_ADDRESS, 0x10);
 
 	return ptr - buf;
 }
@@ -1234,12 +1867,184 @@
 	struct sm501fb_info *info = dev_get_drvdata(dev);
 	char *ptr = buf;
 
-	ptr += sm501fb_show_regs(info, ptr, 0x0, 0x40);
-	ptr += sm501fb_show_regs(info, ptr, SM501_DC_PANEL_HWC_BASE, 0x10);
+	ptr += sm501fb_show_regs(info, ptr, PANEL_DISPLAY_CTRL, 0x40);
+	ptr += sm501fb_show_regs(info, ptr, PANEL_HWC_ADDRESS, 0x10);
 
 	return ptr - buf;
 }
 
+
+
+void sm501fb_copyarea(struct fb_info *info, const struct fb_copyarea *area)
+{
+	struct sm501fb_par *p = (struct sm501fb_par*)info->par;
+	ulong ulreg;
+	if(noca)
+		goto soft;
+ 
+    if (!noaccel)
+    {
+    	if (!area->width || !area->height)
+            return;
+	/*
+       	deCopy(p->screen.sm_addr , 0, info->var.bits_per_pixel,  
+            area->dx, area->dy, area->width, area->height, 
+            p->screen.sm_addr, 0, area->sx, area->sy, 0, 0xC);
+        */
+
+	ulreg=SmRead32(DE_BASE_ADDRESS+DE_STRETCH_FORMAT);	
+	switch (info->var.bits_per_pixel)
+	{
+		case 8:
+			ulreg=FIELD_SET(ulreg,DE_STRETCH_FORMAT,PIXEL_FORMAT,8);
+			break;
+		case 16:
+			ulreg=FIELD_SET(ulreg,DE_STRETCH_FORMAT,PIXEL_FORMAT,16);
+			break;
+		case 24:
+		case 32:		
+			ulreg=FIELD_SET(ulreg,DE_STRETCH_FORMAT,PIXEL_FORMAT,32);
+			break;
+	}
+	SmWrite2D( DE_STRETCH_FORMAT, ulreg);
+        
+	deVideoMem2VideoMemBlt(p->screen.sm_addr, info->fix.line_length,area->sx, area->sy,
+							p->screen.sm_addr,info->fix.line_length,info->var.bits_per_pixel,
+							area->dx,area->dy,area->width,area->height,0xc);
+	
+    }
+    else
+    {
+soft:    
+        cfb_copyarea(info, area);
+    }
+}
+void sm501fb_fillrect(struct fb_info *info, const struct fb_fillrect *rect)
+{
+    	struct sm501fb_par *p = (struct sm501fb_par*)info->par;
+	ulong rop,color,bpp;
+	ulong color32[4]={0xffffff,0xFF0000,0xFF00,0xFF};
+	ulong color16[4]={0x001f,0x07e0,0xf800,0xffff};
+	ulong x,y;
+	
+	
+	if(nofr)
+		goto soft;
+ 
+	if (!noaccel)
+    	{
+       		 if (!rect->width || !rect->height)
+            		return;
+
+		if(info->var.bits_per_pixel == 8)
+			color = rect->color;
+		else
+			color = ((u32*)(info->pseudo_palette))[rect->color];
+	
+		if(rect->rop == ROP_COPY)
+			rop = 0xc;
+		else
+			rop = 0x6;
+		
+	       	
+		deRectFill(p->screen.sm_addr,info->fix.line_length,info->var.bits_per_pixel,
+			 rect->dx, rect->dy, rect->width, rect->height, color,rop);
+		return;
+			
+		if(rect->width != 800 || rect->height != 8)
+		{
+			deRectFill(p->screen.sm_addr,info->fix.line_length,info->var.bits_per_pixel,
+			 rect->dx, rect->dy, rect->width, rect->height, color,rop);
+		}
+		else
+		{
+			gen = gen %4;
+			if(info->var.bits_per_pixel == 16){
+				bpp=16;
+				color=color16[gen];	
+				x=0;
+				y=592;
+			}
+			else{	
+				bpp=32;
+				color=color32[gen];
+				x=400;
+				y=592;
+			}		
+			deRectFill(p->screen.sm_addr,info->fix.line_length,bpp,
+				 x, y, 400,8, color,rop);
+			printk("bpp=%d,color=%08x pos = [%d,%d]\n ",bpp,color,x,y);
+			gen++;
+		}
+	}
+	else
+	{	
+soft:	
+        	cfb_fillrect(info, rect);
+	}
+}
+
+void sm501fb_imageblit(struct fb_info *info, const struct fb_image *image)
+{
+	ulong fgcol,bgcol;	
+	struct sm501fb_par *p = (struct sm501fb_par*)info->par;
+	if(noib)
+		goto soft;
+	
+#if 0
+    {
+	if(sm_accel_busy)
+		sm501fb_Wait_Idle();
+	cfb_imageblit(info, image);
+    }
+#else
+
+	ulong regvalue=SmRead2D(DE_STRETCH_FORMAT);
+	switch(info->var.bits_per_pixel)
+	{
+		case 8:
+			regvalue = FIELD_SET(regvalue,DE_STRETCH_FORMAT,PIXEL_FORMAT,8);
+			break;
+		case 16:
+			regvalue = FIELD_SET(regvalue,DE_STRETCH_FORMAT,PIXEL_FORMAT,16);
+			break;
+		case 32:	
+			regvalue = FIELD_SET(regvalue,DE_STRETCH_FORMAT,PIXEL_FORMAT,32);
+			break;
+	}
+
+	SmWrite2D(DE_STRETCH_FORMAT, regvalue);
+	
+	if(!noaccel && image->depth == 1)		
+	{
+	
+			if(info->fix.visual == FB_VISUAL_TRUECOLOR ||info->fix.visual == FB_VISUAL_DIRECTCOLOR)
+			{
+				fgcol = ((u32*)info->pseudo_palette)[image->fg_color];
+				bgcol = ((u32*)info->pseudo_palette)[image->bg_color];		
+			}
+			else
+			{
+				fgcol = image->fg_color;
+				bgcol = image->bg_color;
+			}
+			deSystemMem2VideoMemMonoBlt(image->data,image->width/8,0,p->screen.sm_addr,
+									info->fix.line_length,info->var.bits_per_pixel,
+									image->dx,image->dy,image->width,image->height,fgcol,bgcol,
+									0xc);		
+		
+	}
+	else
+	{
+soft:	
+		cfb_imageblit(info,image);
+	}
+#endif
+}
+
+
+
+	
 static DEVICE_ATTR(fbregs_pnl, 0444, sm501fb_debug_show_pnl, NULL);
 
 /* framebuffer ops */
@@ -1248,13 +2053,18 @@
 	.owner		= THIS_MODULE,
 	.fb_check_var	= sm501fb_check_var_crt,
 	.fb_set_par	= sm501fb_set_par_crt,
-	.fb_blank	= sm501fb_blank_crt,
+	.fb_blank	= sm501fb_blank,
 	.fb_setcolreg	= sm501fb_setcolreg,
 	.fb_pan_display	= sm501fb_pan_crt,
 	.fb_cursor	= sm501fb_cursor,
-	.fb_fillrect	= cfb_fillrect,
-	.fb_copyarea	= cfb_copyarea,
-	.fb_imageblit	= cfb_imageblit,
+	.fb_fillrect	= sm501fb_fillrect,
+	.fb_copyarea	= sm501fb_copyarea,
+	.fb_imageblit	= sm501fb_imageblit,
+#ifdef __BIG_ENDIAN	
+//	.fb_setcmap = sm501fb_set_cmap,
+	.fb_write = sm501fb_write,
+	.fb_read = sm501fb_read,
+#endif
 };
 
 static struct fb_ops sm501fb_ops_pnl = {
@@ -1262,14 +2072,52 @@
 	.fb_check_var	= sm501fb_check_var_pnl,
 	.fb_set_par	= sm501fb_set_par_pnl,
 	.fb_pan_display	= sm501fb_pan_pnl,
-	.fb_blank	= sm501fb_blank_pnl,
+	.fb_blank	= sm501fb_blank,
 	.fb_setcolreg	= sm501fb_setcolreg,
 	.fb_cursor	= sm501fb_cursor,
-	.fb_fillrect	= cfb_fillrect,
-	.fb_copyarea	= cfb_copyarea,
-	.fb_imageblit	= cfb_imageblit,
+	.fb_fillrect	= sm501fb_fillrect,
+	.fb_copyarea	= sm501fb_copyarea,
+	.fb_imageblit	= sm501fb_imageblit,
+#ifdef __BIG_ENDIAN	
+//	.fb_setcmap = sm501fb_set_cmap,
+	.fb_write = sm501fb_write,
+	.fb_read = sm501fb_read,
+#endif
 };
 
+/* sm501fb_info_alloc
+ *
+ * creates and initialises an sm501fb_info structure
+*/
+
+static struct sm501fb_info *sm501fb_info_alloc(struct fb_info *fbinfo_crt,
+					       struct fb_info *fbinfo_pnl)
+{
+	struct sm501fb_info *info;
+	struct sm501fb_par  *par;
+
+	info = kzalloc(sizeof(struct sm501fb_info), GFP_KERNEL);
+	if (info) {
+		/* set the references back */
+
+		par = fbinfo_crt->par;
+		par->info = info;
+		par->head = HEAD_CRT;
+		fbinfo_crt->pseudo_palette = &par->pseudo_palette;
+
+		par = fbinfo_pnl->par;
+		par->info = info;
+		par->head = HEAD_PANEL;
+		fbinfo_pnl->pseudo_palette = &par->pseudo_palette;
+
+		/* store the two fbs into our info */
+		info->fb[HEAD_CRT] = fbinfo_crt;
+		info->fb[HEAD_PANEL] = fbinfo_pnl;
+	}
+
+	return info;
+}
+
 /* sm501_init_cursor
  *
  * initialise hw cursor parameters
@@ -1277,17 +2125,11 @@
 
 static int sm501_init_cursor(struct fb_info *fbi, unsigned int reg_base)
 {
-	struct sm501fb_par *par;
-	struct sm501fb_info *info;
+	struct sm501fb_par *par = fbi->par;
+	struct sm501fb_info *info = par->info;
 	int ret;
 
-	if (fbi == NULL)
-		return 0;
-
-	par = fbi->par;
-	info = par->info;
-
-	par->cursor_regs = info->regs + reg_base;
+	par->cursor_regs = reg_base;
 
 	ret = sm501_alloc_mem(info, &par->cursor, SM501_MEMF_CURSOR, 1024);
 	if (ret < 0)
@@ -1295,11 +2137,11 @@
 
 	/* initialise the colour registers */
 
-	writel(par->cursor.sm_addr, par->cursor_regs + SM501_OFF_HWC_ADDR);
+	SmWrite32(par->cursor_regs + SM501_OFF_HWC_ADDR, par->cursor.sm_addr);
 
-	writel(0x00, par->cursor_regs + SM501_OFF_HWC_LOC);
-	writel(0x00, par->cursor_regs + SM501_OFF_HWC_COLOR_1_2);
-	writel(0x00, par->cursor_regs + SM501_OFF_HWC_COLOR_3);
+	SmWrite32(par->cursor_regs + SM501_OFF_HWC_LOC, 0x0);
+	SmWrite32(par->cursor_regs + SM501_OFF_HWC_COLOR_1_2, 0x0);
+	SmWrite32(par->cursor_regs + SM501_OFF_HWC_COLOR_3, 0x0);
 	sm501fb_sync_regs(info);
 
 	return 0;
@@ -1314,9 +2156,11 @@
 			 struct platform_device *pdev)
 {
 	struct resource	*res;
-	struct device *dev = &pdev->dev;
-	int k;
-	int ret;
+	struct device *dev;
+	int ret, value;
+
+	info->dev = dev = &pdev->dev;
+	platform_set_drvdata(pdev, info);
 
 	info->irq = ret = platform_get_irq(pdev, 0);
 	if (ret < 0) {
@@ -1324,31 +2168,6 @@
 		dev_warn(dev, "no irq for device\n");
 	}
 
-	/* allocate, reserve and remap resources for registers */
-	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
-	if (res == NULL) {
-		dev_err(dev, "no resource definition for registers\n");
-		ret = -ENOENT;
-		goto err_release;
-	}
-
-	info->regs_res = request_mem_region(res->start,
-					    res->end - res->start,
-					    pdev->name);
-
-	if (info->regs_res == NULL) {
-		dev_err(dev, "cannot claim registers\n");
-		ret = -ENXIO;
-		goto err_release;
-	}
-
-	info->regs = ioremap(res->start, (res->end - res->start)+1);
-	if (info->regs == NULL) {
-		dev_err(dev, "cannot remap registers\n");
-		ret = -ENXIO;
-		goto err_regs_res;
-	}
-
 	/* allocate, reserve resources for framebuffer */
 	res = platform_get_resource(pdev, IORESOURCE_MEM, 2);
 	if (res == NULL) {
@@ -1360,6 +2179,7 @@
 	info->fbmem_res = request_mem_region(res->start,
 					     (res->end - res->start)+1,
 					     pdev->name);
+	
 	if (info->fbmem_res == NULL) {
 		dev_err(dev, "cannot claim framebuffer\n");
 		ret = -ENXIO;
@@ -1373,21 +2193,15 @@
 	}
 
 	info->fbmem_len = (res->end - res->start)+1;
-
-	/* clear framebuffer memory - avoids garbage data on unused fb */
-	memset(info->fbmem, 0, info->fbmem_len);
-
-	/* clear palette ram - undefined at power on */
-	for (k = 0; k < (256 * 3); k++)
-		writel(0, info->regs + SM501_DC_PANEL_PALETTE + (k * 4));
-
+		
 	/* enable display controller */
-	sm501_unit_power(dev->parent, SM501_GATE_DISPLAY, 1);
+	value = SmRead32(CURRENT_GATE);
+	sm501_set_gate(FIELD_SET(value,POWER_MODE0_GATE, DISPLAY, ENABLE) );
 
 	/* setup cursors */
 
-	sm501_init_cursor(info->fb[HEAD_CRT], SM501_DC_CRT_HWC_ADDR);
-	sm501_init_cursor(info->fb[HEAD_PANEL], SM501_DC_PANEL_HWC_ADDR);
+	sm501_init_cursor(info->fb[HEAD_CRT], CRT_HWC_ADDRESS);
+	sm501_init_cursor(info->fb[HEAD_PANEL], PANEL_HWC_ADDRESS);
 
 	return 0; /* everything is setup */
 
@@ -1408,16 +2222,54 @@
 
 static void sm501fb_stop(struct sm501fb_info *info)
 {
+	int value;
 	/* disable display controller */
-	sm501_unit_power(info->dev->parent, SM501_GATE_DISPLAY, 0);
+	value = SmRead32(CURRENT_GATE);
+	sm501_set_gate(FIELD_SET(value,POWER_MODE0_GATE, DISPLAY, DISABLE) );
 
 	iounmap(info->fbmem);
 	release_resource(info->fbmem_res);
 	kfree(info->fbmem_res);
 
-	iounmap(info->regs);
-	release_resource(info->regs_res);
-	kfree(info->regs_res);
+}
+
+static void sm501fb_info_release(struct sm501fb_info *info)
+{
+	kfree(info);
+}
+
+
+void sm501fb_cmap_swap(struct fb_cmap *cmap, struct fb_info *info)
+{
+	int i, start = 0;
+	u16 *red, *green, *blue, *transp;
+	unsigned int val;
+
+	red = cmap->red;
+	green = cmap->green;
+	blue = cmap->blue;
+	transp = cmap->transp;
+	start = cmap->start;
+
+		for (i = 0; i < cmap->len; i++) {
+			val  = chan_to_field(red[i],   &info->var.red);
+			val |= chan_to_field(green[i], &info->var.green);
+			val |= chan_to_field(blue[i],  &info->var.blue);
+		
+			if (transp){
+				val |= chan_to_field(transp[i],  &info->var.transp);
+			}
+			val = (val& 0xff00ff00>>8)|(val& 0x00ff00ff<<8);
+
+			red[i] = field_to_chan(val,&info->var.red);
+			green[i] = field_to_chan(val,&info->var.green);
+			blue[i] = field_to_chan(val,&info->var.blue);
+
+			if (transp+i)
+				transp[i] |= field_to_chan(val,&info->var.transp);
+		}
+
+	return ;
 }
 
 static int sm501fb_init_fb(struct fb_info *fb,
@@ -1430,25 +2282,31 @@
 	unsigned long ctrl;
 	unsigned int enable;
 	int ret;
-
+	
+	
 	switch (head) {
 	case HEAD_CRT:
-		pd = info->pdata->fb_crt;
-		ctrl = readl(info->regs + SM501_DC_CRT_CONTROL);
-		enable = (ctrl & SM501_DC_CRT_CONTROL_ENABLE) ? 1 : 0;
 
+		pd = info->pdata->fb_crt;
+		ctrl = SmRead32(CRT_DISPLAY_CTRL);
+		enable = FIELD_GET(ctrl, CRT_DISPLAY_CTRL, PLANE);
 		/* ensure we set the correct source register */
 		if (info->pdata->fb_route != SM501_FB_CRT_PANEL) {
-			ctrl |= SM501_DC_CRT_CONTROL_SEL;
-			writel(ctrl, info->regs + SM501_DC_CRT_CONTROL);
+#ifdef CONFIG_FB_SM501_DUAL_HEAD
+		printk("monk config_fb_sm501_dual_head \n");
+		ctrl = FIELD_SET(ctrl, CRT_DISPLAY_CTRL, SELECT, CRT);
+#else
+		ctrl = FIELD_SET(ctrl, CRT_DISPLAY_CTRL, SELECT, PANEL/*CRT*/);
+#endif
+			SmWrite32(CRT_DISPLAY_CTRL, ctrl);
 		}
-
 		break;
 
 	case HEAD_PANEL:
 		pd = info->pdata->fb_pnl;
-		ctrl = readl(info->regs + SM501_DC_PANEL_CONTROL);
-		enable = (ctrl & SM501_DC_PANEL_CONTROL_EN) ? 1 : 0;
+		ctrl = SmRead32(PANEL_DISPLAY_CTRL);
+		enable =FIELD_GET(ctrl, PANEL_DISPLAY_CTRL, PLANE);
+
 		break;
 
 	default:
@@ -1458,14 +2316,18 @@
 		BUG();
 	}
 
+
+
 	dev_info(info->dev, "fb %s %sabled at start\n",
 		 fbname, enable ? "en" : "dis");
 
 	/* check to see if our routing allows this */
 
+
+	
 	if (head == HEAD_CRT && info->pdata->fb_route == SM501_FB_CRT_PANEL) {
-		ctrl &= ~SM501_DC_CRT_CONTROL_SEL;
-		writel(ctrl, info->regs + SM501_DC_CRT_CONTROL);
+		ctrl = FIELD_SET(ctrl, CRT_DISPLAY_CTRL, SELECT, PANEL);
+		SmWrite32(CRT_DISPLAY_CTRL, ctrl);
 		enable = 0;
 	}
 
@@ -1474,12 +2336,12 @@
 	memcpy(&par->ops,
 	       (head == HEAD_CRT) ? &sm501fb_ops_crt : &sm501fb_ops_pnl,
 	       sizeof(struct fb_ops));
-
+		
 	/* update ops dependant on what we've been passed */
 
 	if ((pd->flags & SM501FB_FLAG_USE_HWCURSOR) == 0)
 		par->ops.fb_cursor = NULL;
-
+	
 	fb->fbops = &par->ops;
 	fb->flags = FBINFO_FLAG_DEFAULT |
 		FBINFO_HWACCEL_XPAN | FBINFO_HWACCEL_YPAN;
@@ -1495,15 +2357,20 @@
 
 	/* screenmode */
 
-	fb->var.nonstd		= 0;
+	if (dcolor)
+		fb->var.nonstd = 1;
 	fb->var.activate	= FB_ACTIVATE_NOW;
 	fb->var.accel_flags	= 0;
 	fb->var.vmode		= FB_VMODE_NONINTERLACED;
 	fb->var.bits_per_pixel  = 16;
 
-	if (enable && (pd->flags & SM501FB_FLAG_USE_INIT_MODE) && 0) {
-		/* TODO read the mode from the current display */
-
+	if (enable && (pd->flags & SM501FB_FLAG_USE_INIT_MODE)&&0) {
+/*		fb->var.xres = screen_info.lfb_width;
+		fb->var.yres = screen_info.lfb_height;
+		fb->var.bits_per_pixel = screen_info.lfb_depth;
+		fb->var.xres_virtual = fb->var.xres;
+		fb->var.yres_virtual = fb->var.yres;
+*/
 	} else {
 		if (pd->def_mode) {
 			dev_info(info->dev, "using supplied mode\n");
@@ -1513,9 +2380,11 @@
 			fb->var.xres_virtual = fb->var.xres;
 			fb->var.yres_virtual = fb->var.yres;
 		} else {
-			ret = fb_find_mode(&fb->var, fb,
-					   NULL, NULL, 0, NULL, 8);
-
+			ret = fb_find_mode(&fb->var, fb, mode_option, SM501modedb, ARRAY_SIZE(SM501modedb), NULL, 16);
+			if(ret != 1)
+			{
+			ret = fb_find_mode(&fb->var, fb, mode_option, NULL, 0, NULL, 16);
+			}
 			if (ret == 0 || ret == 4) {
 				dev_err(info->dev,
 					"failed to get initial mode\n");
@@ -1523,9 +2392,10 @@
 			}
 		}
 	}
-
+	
 	/* initialise and set the palette */
 	fb_alloc_cmap(&fb->cmap, NR_PALETTE, 0);
+//	sm501fb_cmap_swap(&fb->cmap, fb);
 	fb_set_cmap(&fb->cmap, fb);
 
 	ret = (fb->fbops->fb_check_var)(&fb->var, fb);
@@ -1540,7 +2410,14 @@
 
 /* default platform data if none is supplied (ie, PCI device) */
 
+static struct fb_videomode default_mode = {
+	/* 640x480 @ 60 Hz, 31.5 kHz hsync */
+	NULL, 60, 640, 480, 39721, 40, 24, 32, 11, 96, 2,
+	0, FB_VMODE_NONINTERLACED
+};
+
 static struct sm501_platdata_fbsub sm501fb_pdata_crt = {
+//	.def_mode	= &default_mode;
 	.flags		= (SM501FB_FLAG_USE_INIT_MODE |
 			   SM501FB_FLAG_USE_HWCURSOR |
 			   SM501FB_FLAG_USE_HWACCEL |
@@ -1549,6 +2426,7 @@
 };
 
 static struct sm501_platdata_fbsub sm501fb_pdata_pnl = {
+//	.def_mode	= &default_mode;
 	.flags		= (SM501FB_FLAG_USE_INIT_MODE |
 			   SM501FB_FLAG_USE_HWCURSOR |
 			   SM501FB_FLAG_USE_HWACCEL |
@@ -1564,185 +2442,150 @@
 static char driver_name_crt[] = "sm501fb-crt";
 static char driver_name_pnl[] = "sm501fb-panel";
 
-static int __devinit sm501fb_probe_one(struct sm501fb_info *info,
-				       enum sm501_controller head)
+static int __init sm501fb_probe(struct platform_device *pdev)
 {
-	unsigned char *name = (head == HEAD_CRT) ? "crt" : "panel";
-	struct sm501_platdata_fbsub *pd;
-	struct sm501fb_par *par;
-	struct fb_info *fbi;
-
-	pd = (head == HEAD_CRT) ? info->pdata->fb_crt : info->pdata->fb_pnl;
+	struct sm501fb_info *info;
+	struct device	    *dev = &pdev->dev;
+	struct fb_info	    *fbinfo_crt;
+	struct fb_info	    *fbinfo_pnl;
+	int		     ret;
 
-	/* Do not initialise if we've not been given any platform data */
-	if (pd == NULL) {
-		dev_info(info->dev, "no data for fb %s (disabled)\n", name);
-		return 0;
-	}
+	/* allocate our framebuffers */
 
-	fbi = framebuffer_alloc(sizeof(struct sm501fb_par), info->dev);
-	if (fbi == NULL) {
-		dev_err(info->dev, "cannot allocate %s framebuffer\n", name);
+	fbinfo_crt = framebuffer_alloc(sizeof(struct sm501fb_par), dev);
+	if (fbinfo_crt == NULL) {
+		dev_err(dev, "cannot allocate crt framebuffer\n");
 		return -ENOMEM;
 	}
 
-	par = fbi->par;
-	par->info = info;
-	par->head = head;
-	fbi->pseudo_palette = &par->pseudo_palette;
-
-	info->fb[head] = fbi;
-
-	return 0;
-}
-
-/* Free up anything allocated by sm501fb_init_fb */
-
-static void sm501_free_init_fb(struct sm501fb_info *info,
-				enum sm501_controller head)
-{
-	struct fb_info *fbi = info->fb[head];
-
-	fb_dealloc_cmap(&fbi->cmap);
-}
-
-static int __devinit sm501fb_start_one(struct sm501fb_info *info,
-				       enum sm501_controller head,
-				       const char *drvname)
-{
-	struct fb_info *fbi = info->fb[head];
-	int ret;
-
-	if (!fbi)
-		return 0;
-
-	ret = sm501fb_init_fb(info->fb[head], head, drvname);
-	if (ret) {
-		dev_err(info->dev, "cannot initialise fb %s\n", drvname);
-		return ret;
+	fbinfo_pnl = framebuffer_alloc(sizeof(struct sm501fb_par), dev);
+	if (fbinfo_pnl == NULL) {
+		dev_err(dev, "cannot allocate panel framebuffer\n");
+		ret = -ENOMEM;
+		goto fbinfo_crt_alloc_fail;
 	}
 
-	ret = register_framebuffer(info->fb[head]);
-	if (ret) {
-		dev_err(info->dev, "failed to register fb %s\n", drvname);
-		sm501_free_init_fb(info, head);
-		return ret;
+	info = sm501fb_info_alloc(fbinfo_crt, fbinfo_pnl);
+	if (info == NULL) {
+		dev_err(dev, "cannot allocate par\n");
+		ret = -ENOMEM;
+		goto sm501fb_alloc_fail;
 	}
 
-	dev_info(info->dev, "fb%d: %s frame buffer\n", fbi->node, fbi->fix.id);
-
-	return 0;
-}
-
-static int __devinit sm501fb_probe(struct platform_device *pdev)
-{
-	struct sm501fb_info *info;
-	struct device *dev = &pdev->dev;
-	int ret;
-
-	/* allocate our framebuffers */
-
-	info = kzalloc(sizeof(struct sm501fb_info), GFP_KERNEL);
-	if (!info) {
-		dev_err(dev, "failed to allocate state\n");
-		return -ENOMEM;
-	}
-
-	info->dev = dev = &pdev->dev;
-	platform_set_drvdata(pdev, info);
-
 	if (dev->parent->platform_data) {
 		struct sm501_platdata *pd = dev->parent->platform_data;
 		info->pdata = pd->fb;
 	}
 
 	if (info->pdata == NULL) {
-		dev_info(dev, "using default configuration data\n");
+		dev_dbg(dev, "using default configuration data\n");
 		info->pdata = &sm501fb_def_pdata;
 	}
 
-	/* probe for the presence of each panel */
+	if (!nomtrr)
+		sm501fb_set_mtrr(info->dev->parent);
+	
+	/* start the framebuffers */
 
-	ret = sm501fb_probe_one(info, HEAD_CRT);
-	if (ret < 0) {
-		dev_err(dev, "failed to probe CRT\n");
-		goto err_alloc;
+	ret = sm501fb_start(info, pdev);
+	if (ret) {
+		dev_err(dev, "cannot initialise SM501\n");
+		goto sm501fb_start_fail;
 	}
 
-	ret = sm501fb_probe_one(info, HEAD_PANEL);
-	if (ret < 0) {
-		dev_err(dev, "failed to probe PANEL\n");
-		goto err_probed_crt;
-	}
 
-	if (info->fb[HEAD_PANEL] == NULL &&
-	    info->fb[HEAD_CRT] == NULL) {
-		dev_err(dev, "no framebuffers found\n");
-		goto err_alloc;
+	memset(info->fbmem,0,info->fbmem_len); //Teddy:Clease screen before set timing registers.
+	
+	/* CRT framebuffer setup */
+
+	ret = sm501fb_init_fb(fbinfo_crt, HEAD_CRT, driver_name_crt);
+	if (ret) {
+		dev_err(dev, "cannot initialise CRT fb\n");
+		goto sm501fb_start_fail;
 	}
 
-	/* get the resources for both of the framebuffers */
+	/* Panel framebuffer setup */
 
-	ret = sm501fb_start(info, pdev);
+	ret = sm501fb_init_fb(fbinfo_pnl, HEAD_PANEL, driver_name_pnl);
 	if (ret) {
-		dev_err(dev, "cannot initialise SM501\n");
-		goto err_probed_panel;
+		dev_err(dev, "cannot initialise Panel fb\n");
+		goto sm501fb_start_fail;
 	}
 
-	ret = sm501fb_start_one(info, HEAD_CRT, driver_name_crt);
-	if (ret) {
-		dev_err(dev, "failed to start CRT\n");
-		goto err_started;
+	/* Init sm501 drawing engine */
+	if(!noaccel)
+       	{
+       		//deInit(fbinfo_crt->var.xres, fbinfo_crt->var.yres, fbinfo_crt->var.bits_per_pixel); 
+		printk("2d accel enabled\n");			
+	}
+	else
+	{
+		printk("2d accel disabled\n");
 	}
 
-	ret = sm501fb_start_one(info, HEAD_PANEL, driver_name_pnl);
-	if (ret) {
-		dev_err(dev, "failed to start Panel\n");
-		goto err_started_crt;
+	
+	
+	
+	/* register framebuffers */
+
+	ret = register_framebuffer(fbinfo_pnl);
+	if (ret < 0) {
+		dev_err(dev, "failed to register panel fb (%d)\n", ret);
+		goto register_pnl_fail;
 	}
 
-	/* create device files */
+	ret = register_framebuffer(fbinfo_crt);
+	if (ret < 0) {
+		dev_err(dev, "failed to register CRT fb (%d)\n", ret);
+		goto register_crt_fail;
+	}
+	
+
+	dev_dbg(dev, "accel: x:%dx%d bit_pixel: %d\n", fbinfo_crt->var.xres, fbinfo_crt->var.yres, fbinfo_crt->var.bits_per_pixel);	
+	
+	dev_dbg(dev, "fb%d: %s frame buffer device\n",
+		 fbinfo_crt->node, fbinfo_crt->fix.id);
+
+	dev_dbg(dev, "fb%d: %s frame buffer device\n",
+	       fbinfo_pnl->node, fbinfo_pnl->fix.id);
 
+	/* create device files */
 	ret = device_create_file(dev, &dev_attr_crt_src);
 	if (ret)
-		goto err_started_panel;
-
+		goto crtsrc_fail;
 	ret = device_create_file(dev, &dev_attr_fbregs_pnl);
 	if (ret)
-		goto err_attached_crtsrc_file;
-
+		goto fbregs_pnl_fail;
 	ret = device_create_file(dev, &dev_attr_fbregs_crt);
 	if (ret)
-		goto err_attached_pnlregs_file;
+		goto fbregs_crt_fail;
 
 	/* we registered, return ok */
 	return 0;
 
-err_attached_pnlregs_file:
+ fbregs_crt_fail:
 	device_remove_file(dev, &dev_attr_fbregs_pnl);
 
-err_attached_crtsrc_file:
+ fbregs_pnl_fail:
 	device_remove_file(dev, &dev_attr_crt_src);
 
-err_started_panel:
-	unregister_framebuffer(info->fb[HEAD_PANEL]);
-	sm501_free_init_fb(info, HEAD_PANEL);
-
-err_started_crt:
-	unregister_framebuffer(info->fb[HEAD_CRT]);
-	sm501_free_init_fb(info, HEAD_CRT);
+ crtsrc_fail:
+	unregister_framebuffer(fbinfo_pnl);
+
+ register_pnl_fail:
+	unregister_framebuffer(fbinfo_crt);
 
-err_started:
+ register_crt_fail:
 	sm501fb_stop(info);
 
-err_probed_panel:
-	framebuffer_release(info->fb[HEAD_PANEL]);
+ sm501fb_start_fail:
+	sm501fb_info_release(info);
 
-err_probed_crt:
-	framebuffer_release(info->fb[HEAD_CRT]);
+ sm501fb_alloc_fail:
+	framebuffer_release(fbinfo_pnl);
 
-err_alloc:
-	kfree(info);
+ fbinfo_crt_alloc_fail:
+	framebuffer_release(fbinfo_crt);
 
 	return ret;
 }
@@ -1761,14 +2604,12 @@
 	device_remove_file(&pdev->dev, &dev_attr_fbregs_pnl);
 	device_remove_file(&pdev->dev, &dev_attr_crt_src);
 
-	sm501_free_init_fb(info, HEAD_CRT);
-	sm501_free_init_fb(info, HEAD_PANEL);
-
-	unregister_framebuffer(fbinfo_crt);
 	unregister_framebuffer(fbinfo_pnl);
-
+	unregister_framebuffer(fbinfo_crt);
+	if (!nomtrr)
+		sm501fb_unset_mtrr(info->dev->parent);
 	sm501fb_stop(info);
-	kfree(info);
+	sm501fb_info_release(info);
 
 	framebuffer_release(fbinfo_pnl);
 	framebuffer_release(fbinfo_crt);
@@ -1795,7 +2636,6 @@
 	acquire_console_sem();
 	fb_set_suspend(fbi, 1);
 	release_console_sem();
-
 	/* backup copies in case chip is powered down over suspend */
 
 	par->store_fb = vmalloc(par->screen.size);
@@ -1810,9 +2650,6 @@
 		goto err_nocursor;
 	}
 
-	dev_dbg(info->dev, "suspending screen to %p\n", par->store_fb);
-	dev_dbg(info->dev, "suspending cursor to %p\n", par->store_cursor);
-
 	memcpy_fromio(par->store_fb, par->screen.k_addr, par->screen.size);
 	memcpy_fromio(par->store_cursor, par->cursor.k_addr, par->cursor.size);
 
@@ -1866,36 +2703,28 @@
 {
 	struct sm501fb_info *info = platform_get_drvdata(pdev);
 
-	/* store crt control to resume with */
-	info->pm_crt_ctrl = readl(info->regs + SM501_DC_CRT_CONTROL);
-
 	sm501fb_suspend_fb(info, HEAD_CRT);
 	sm501fb_suspend_fb(info, HEAD_PANEL);
 
 	/* turn off the clocks, in case the device is not powered down */
-	sm501_unit_power(info->dev->parent, SM501_GATE_DISPLAY, 0);
+
+	int value = SmRead32(CURRENT_GATE);
+	sm501_set_gate(FIELD_SET(value,POWER_MODE0_GATE, DISPLAY,DISABLE)&
+		FIELD_SET(value, POWER_MODE0_GATE, 2D, DISABLE)&
+		FIELD_SET(value, POWER_MODE0_GATE, CSC, DISABLE) );
 
 	return 0;
 }
 
-#define SM501_CRT_CTRL_SAVE (SM501_DC_CRT_CONTROL_TVP |        \
-			     SM501_DC_CRT_CONTROL_SEL)
-
-
 static int sm501fb_resume(struct platform_device *pdev)
 {
 	struct sm501fb_info *info = platform_get_drvdata(pdev);
-	unsigned long crt_ctrl;
-
-	sm501_unit_power(info->dev->parent, SM501_GATE_DISPLAY, 1);
-
-	/* restore the items we want to be saved for crt control */
-
-	crt_ctrl = readl(info->regs + SM501_DC_CRT_CONTROL);
-	crt_ctrl &= ~SM501_CRT_CTRL_SAVE;
-	crt_ctrl |= info->pm_crt_ctrl & SM501_CRT_CTRL_SAVE;
-	writel(crt_ctrl, info->regs + SM501_DC_CRT_CONTROL);
 
+	int value = SmRead32(CURRENT_GATE);
+	sm501_set_gate(FIELD_SET(value, POWER_MODE0_GATE, DISPLAY, ENABLE)|
+		FIELD_SET(value, POWER_MODE0_GATE, 2D, ENABLE)|
+		FIELD_SET(value, POWER_MODE0_GATE, CSC, ENABLE) );
+	
 	sm501fb_resume_fb(info, HEAD_CRT);
 	sm501fb_resume_fb(info, HEAD_PANEL);
 
@@ -1910,7 +2739,7 @@
 static struct platform_driver sm501fb_driver = {
 	.probe		= sm501fb_probe,
 	.remove		= sm501fb_remove,
-	.suspend	= sm501fb_suspend,
+	.suspend		= sm501fb_suspend,
 	.resume		= sm501fb_resume,
 	.driver		= {
 		.name	= "sm501-fb",
@@ -1918,8 +2747,54 @@
 	},
 };
 
+/*
+ * Interface to the world
+ */
+static int  __devinit sm501fb_setup(char *options)
+{
+	char *this_opt;
+
+	if (!options || !*options)
+		return 0;
+
+	while ((this_opt = strsep(&options, ",")) != NULL) {
+		if (!*this_opt) continue;
+
+		smdbg("option %s\n", this_opt);
+
+		if (!strncmp(this_opt,"noaccel",7))
+			noaccel = 1;
+		else	if (!strncmp(this_opt,"dcolor",6))
+			dcolor = 1;
+		else if(!strncmp(this_opt,"nofr",4))
+			nofr=1;
+		else if(!strncmp(this_opt,"noca",4))
+			noca=1;
+		else if(!strncmp(this_opt,"noib",4))
+			noib=1;
+		else
+			mode_option = this_opt;
+		//noaccel = 0;
+	}
+	return 0;
+}
+
+
 static int __devinit sm501fb_init(void)
 {
+	char *option = NULL;
+
+        /* Jason (08/05/2009):
+         * if there are params passed using modprobe, use those
+         * if not, use the booting vesa params.
+         */
+#ifdef MODULE
+        option = mode_option;
+#else
+        if (fb_get_options("sm501fb", &option))
+            return -ENODEV;
+#endif
+	sm501fb_setup(option);
 	return platform_driver_register(&sm501fb_driver);
 }
 
@@ -1931,6 +2806,18 @@
 module_init(sm501fb_init);
 module_exit(sm501fb_cleanup);
 
-MODULE_AUTHOR("Ben Dooks, Vincent Sanders");
-MODULE_DESCRIPTION("SM501 Framebuffer driver");
+module_param(mode_option, charp, 0);
+MODULE_PARM_DESC(mode_option, "Initial mode (default=" DEFAULT_MODE ")");
+module_param(noaccel, bool, 0);
+MODULE_PARM_DESC(noaccel, "bool: Disable acceleration support (0 or 1=disabled) (default=0)");
+module_param(mirror, bool, 0);
+MODULE_PARM_DESC(mirror, "bool: Mirror the display to both monitors");
+module_param(dcolor, bool, 0);
+MODULE_PARM_DESC(dcolor, "bool: Enable Direct color.");
+
+module_param(nomtrr, bool, 1);
+MODULE_PARM_DESC(nomtrr, "bool: disable use of MTRR registers");
+
+MODULE_AUTHOR("Boyod Yang, Ben Dooks, Vincent Sanders");
+MODULE_DESCRIPTION("SM50x Framebuffer driver");
 MODULE_LICENSE("GPL v2");
